# coding: utf-8
import re
import time

from .baseview import BaseView


API_MAP = dict(start='schedule', stop='cancel', forcestop='cancel', liststats='logs/stats')


class ApiView(BaseView):

    def __init__(self):
        super(ApiView, self).__init__()

        self.opt = self.view_args['opt']
        self.project = self.view_args['project']
        self.version_spider_job = self.view_args['version_spider_job']

        self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVER, API_MAP.get(self.opt, self.opt))
        self.data = None
        self.status_code = 0
        self.js = {}

    def dispatch_request(self, **kwargs):
        self.update_url()
        self.update_data()
        self.get_result()
        self.handle_result()
        return self.json_dumps(self.js, sort_keys=False, as_response=True)

    def update_url(self):
        if self.opt in ['listversions', 'listjobs']:
            self.url += '?project=%s' % self.project
        elif self.opt == 'listspiders':
            if self.version_spider_job == self.DEFAULT_LATEST_VERSION:
                self.url += '?project=%s' % self.project
            else:
                # Should be _version
                self.url += '?project=%s&_version=%s' % (self.project, self.version_spider_job)

    def update_data(self):
        self.data = dict(project=self.project)
        if self.opt == 'start':
            self.data['spider'] = self.version_spider_job
            self.data['jobid'] = self.get_now_string()
        elif self.opt in ['stop', 'forcestop']:
            self.data['job'] = self.version_spider_job
        elif self.opt == 'delversion':
            self.data['version'] = self.version_spider_job
        elif self.opt == 'delproject':
            pass
        else:
            self.data = None

    def get_result(self):
        timeout = 3 if self.opt == 'daemonstatus' else 60
        dumps_json = self.opt not in ['daemonstatus', 'liststats']
        times = 2 if self.opt == 'forcestop' else 1
        for __ in range(times):
            self.status_code, self.js = self.make_request(self.url, data=self.data, auth=self.AUTH,
                                                          as_json=True, dumps_json=dumps_json, timeout=timeout)
            if times != 1:
                self.js['times'] = times
                time.sleep(2)

    def handle_result(self):
        if self.status_code != 200:
            if self.opt == 'liststats':
                if self.project and self.version_spider_job:  # 'List Stats' in the Servers page
                    if self.status_code == 404:
                        self.js = dict(status=self.OK, tip="'pip install logparser' and run command 'logparser'")
                else:  # XMLHttpRequest in the Jobs page; POST in jobs.py
                    self.js['tip'] = ("'pip install logparser' on host '%s' and run command 'logparser' "
                                      "to show crawled_pages and scraped_items. ") % self.SCRAPYD_SERVER
            else:
                self.js['tip'] = "Make sure that your Scrapyd server is accessable. "
        elif self.js['status'] != self.OK:
            if re.search('No such file|no active project', self.js.get('message', '')):
                self.js['tip'] = "Maybe the project had been deleted, check out the Projects page. "
            elif self.opt == 'listversions':
                self.js['tip'] = (
                    "Maybe it's caused by failing to compare versions, "
                    "you can check out the HELP section in the Deploy Project page for more info, "
                    "and solve the problem in the Projects page. "
                )
            elif self.opt == 'listspiders' and re.search("TypeError: 'tuple'", self.js.get('message', '')):
                self.js['tip'] = "Maybe it's a broken project, check out the Projects page to delete it. "
        elif self.opt == 'liststats':
            if self.js.get('logparser_version') != self.LOGPARSER_VERSION:
                if self.project and self.version_spider_job:  # 'List Stats' in the Servers page
                    tip = "'pip install --upgrade logparser' to update LogParser to v%s" % self.LOGPARSER_VERSION
                    self.js = dict(status=self.OK, tip=tip)
                else:  # XMLHttpRequest in the Jobs page; POST in jobs.py
                    self.js['tip'] = ("'pip install --upgrade logparser' on host '%s' and run command 'logparser' "
                                      "to update LogParser to v%s") % (self.SCRAPYD_SERVER, self.LOGPARSER_VERSION)
                    self.js['status'] = self.ERROR
            elif self.project and self.version_spider_job:  # 'List Stats' in the Servers page
                self.extract_pages_items()

    def extract_pages_items(self):
        details = None
        if self.project in self.js['datas']:
            for spider in self.js['datas'][self.project]:
                for jobid in self.js['datas'][self.project][spider]:
                    if jobid == self.version_spider_job:
                        details = self.js['datas'][self.project][spider][self.version_spider_job]
                        self.js['project'] = self.project
                        self.js['spider'] = spider
                        self.js['jobid'] = jobid
                        break
        if not details:
            details = dict(pages=self.NA, items=self.NA)
        details.setdefault('project', self.project)
        details.setdefault('spider', self.NA)
        details.setdefault('jobid', self.version_spider_job)
        details['logparser_version'] = self.js.get('logparser_version', None)
        self.js = dict(status=self.OK, details=details)



# coding: utf-8
import atexit
import logging
from pprint import pformat

from apscheduler.events import EVENT_JOB_MAX_INSTANCES, EVENT_JOB_REMOVED
from apscheduler.executors.pool import ThreadPoolExecutor  # , ProcessPoolExecutor
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.schedulers.background import BackgroundScheduler

from ..common import handle_metadata
from ..vars import APSCHEDULER_DATABASE_URI, TIMER_TASKS_HISTORY_LOG


apscheduler_logger = logging.getLogger('apscheduler')
# _handler = logging.StreamHandler()
# logging.FileHandler(filename, mode='a', encoding=None, delay=False)
_handler = logging.FileHandler(TIMER_TASKS_HISTORY_LOG, mode='a', encoding='utf-8')
_handler.setLevel(logging.WARNING)
_formatter = logging.Formatter(fmt="[%(asctime)s] %(levelname)s in %(name)s: %(message)s")
_handler.setFormatter(_formatter)
apscheduler_logger.addHandler(_handler)


# EVENT_JOB_REMOVED = 2 ** 10
# {'alias': None, 'code': 1024, 'job_id': '1', 'jobstore': 'default'}
# EVENT_JOB_MAX_INSTANCES = 2 ** 16
EVENT_MAP = {EVENT_JOB_MAX_INSTANCES: 'EVENT_JOB_MAX_INSTANCES', EVENT_JOB_REMOVED: 'EVENT_JOB_REMOVED'}

jobstores = {
    'default': SQLAlchemyJobStore(url=APSCHEDULER_DATABASE_URI),
    'memory': MemoryJobStore()
}
executors = {
    'default': ThreadPoolExecutor(20),
    # 'processpool': ProcessPoolExecutor(5)
}
job_defaults = {
    'coalesce': True,
    'max_instances': 1
}
# https://apscheduler.readthedocs.io/en/latest/userguide.html
# scheduler = BackgroundScheduler(jobstores=jobstores, executors=executors, job_defaults=job_defaults, timezone=utc)
scheduler = BackgroundScheduler(jobstores=jobstores, executors=executors, job_defaults=job_defaults)


# https://apscheduler.readthedocs.io/en/latest/userguide.html#scheduler-events
# EVENT_JOB_EXECUTED: 'code': 4096, 'exception': None
# EVENT_JOB_ERROR: 'code': 8192, 'exception': xxx
# apscheduler/executors/base.py
    # events.append(JobExecutionEvent(EVENT_JOB_MISSED, job.id, jobstore_alias,
    #                                 run_time))
    # logger.warning('Run time of job "%s" was missed by %s', job, difference)
# WARNING in apscheduler.executors.default: Run time of job "task_1" was missed by 0:00:26.030600
# apscheduler/schedulers/base.py
#     self._logger = maybe_ref(config.pop('logger', None)) or getLogger('apscheduler.scheduler')
#     self._logger.warning(
#         'Execution of job "%s" skipped: maximum number of running '
#         'instances reached (%d)', job, job.max_instances)
#     event = JobSubmissionEvent(EVENT_JOB_MAX_INSTANCES, job.id,
#                                jobstore_alias, run_times)
#     events.append(event)

# EVENT_JOB_MAX_INSTANCES: 'job_id': 'jobs_snapshot', 'jobstore': 'memory',
# WARNING in apscheduler.scheduler: Execution of job "create_jobs_snapshot (trigger: interval[0:00:10],
# next run at: " skipped: maximum number of running instances reached (1)
def my_listener(event):
    msg = "%s: \n%s\n" % (EVENT_MAP[event.code], pformat(vars(event), indent=4))
    # logger defined outside the callback of add_listener does not take effect?!
    # In case JOBS_SNAPSHOT_INTERVAL is set too short, like 10 seconds
    if event.jobstore != 'default':
        logging.getLogger('apscheduler').info(msg)
    else:
        logging.getLogger('apscheduler').warning(msg)

    # if hasattr(event, 'exception') and event.exception:
    #     print(event.exception)
    # if hasattr(event, 'traceback') and event.traceback:
    #     print(event.traceback)


# To trigger EVENT_JOB_MAX_INSTANCES:
#   add sleep in execute_task()
#   second: */10, max_instances: 2
# EVENT_JOB_ERROR and EVENT_JOB_MISSED are caught by logging.FileHandler
scheduler.add_listener(my_listener, EVENT_JOB_MAX_INSTANCES | EVENT_JOB_REMOVED)

# if scheduler.state == STATE_STOPPED:
scheduler.start(paused=True)


def shutdown_scheduler():
    apscheduler_logger.debug("Scheduled tasks: %s", scheduler.get_jobs())
    apscheduler_logger.warning("Shutting down the scheduler for timer tasks gracefully, "
                               "wait until all currently executing tasks are finished")
    apscheduler_logger.warning("The main pid is %s. Kill it manually if you don't want to wait",
                               handle_metadata().get('main_pid'))
    scheduler.shutdown()
    # apscheduler_logger.info("Waits until all currently executing jobs are finished. "
    #                         "Press CTRL+C to force unclean shutdown")
    # try:
    #     scheduler.shutdown()
    # except KeyboardInterrupt:
    #     apscheduler_logger.warning("Forcing unclean shutdown")
    #     scheduler.shutdown(wait=False)
    # apscheduler_logger.info("Good Bye")


# https://stackoverflow.com/questions/21214270/scheduling-a-function-to-run-every-hour-on-flask
atexit.register(lambda: shutdown_scheduler())



# coding: utf-8
from datetime import datetime
from pprint import pformat
import time

from flask_sqlalchemy import SQLAlchemy

from .vars import STATE_RUNNING


db = SQLAlchemy(session_options=dict(autocommit=False, autoflush=True))


# TODO: Database Migrations https://blog.miguelgrinberg.com/post/the-flask-mega-tutorial-part-iv-database
# http://flask-sqlalchemy.pocoo.org/2.3/binds/#binds
class Metadata(db.Model):
    __tablename__ = 'metadata'
    __bind_key__ = 'metadata'

    id = db.Column(db.Integer, primary_key=True)
    version = db.Column(db.String(20), unique=True, nullable=False)
    last_check_update_timestamp = db.Column(db.Float, unique=False, default=time.time)
    main_pid = db.Column(db.Integer, unique=False, nullable=True)
    logparser_pid = db.Column(db.Integer, unique=False, nullable=True)
    poll_pid = db.Column(db.Integer, unique=False, nullable=True)
    pageview = db.Column(db.Integer, unique=False, nullable=False, default=0)
    url_scrapydweb = db.Column(db.Text(), unique=False, nullable=False, default='http://127.0.0.1:5000')
    url_jobs = db.Column(db.String(255), unique=False, nullable=False, default='/1/jobs/')
    url_schedule_task = db.Column(db.String(255), unique=False, nullable=False, default='/1/schedule/task/')
    url_delete_task_result = db.Column(db.String(255), unique=False, nullable=False, default='/1/tasks/xhr/delete/1/1/')
    username = db.Column(db.String(255), unique=False, nullable=True)
    password = db.Column(db.String(255), unique=False, nullable=True)
    scheduler_state = db.Column(db.Integer, unique=False, nullable=False, default=STATE_RUNNING)
    jobs_per_page = db.Column(db.Integer, unique=False, nullable=False, default=100)
    tasks_per_page = db.Column(db.Integer, unique=False, nullable=False, default=100)
    jobs_style = db.Column(db.String(8), unique=False, nullable=False, default='database')  # 'classic'

    def __repr__(self):
        return pformat(vars(self))


# TODO: Timezone Conversions https://blog.miguelgrinberg.com/post/the-flask-mega-tutorial-part-xii-dates-and-times
def create_jobs_table(server):
    class Job(db.Model):
        __tablename__ = server
        __bind_key__ = 'jobs'
        # https://stackoverflow.com/questions/10059345/sqlalchemy-unique-across-multiple-columns
        # https://stackoverflow.com/questions/43975349/why-uniqueconstraint-doesnt-work-in-flask-sqlalchemy
        __table_args__ = (db.UniqueConstraint('project', 'spider', 'job'), )

        id = db.Column(db.Integer, primary_key=True)
        project = db.Column(db.String(255), unique=False, nullable=False)  # Pending
        spider = db.Column(db.String(255), unique=False, nullable=False)  # Pending
        job = db.Column(db.String(255), unique=False, nullable=False)  # Pending
        status = db.Column(db.String(1), unique=False, nullable=False, index=True)  # Pending 0, Running 1, Finished 2
        deleted = db.Column(db.String(1), unique=False, nullable=False, default='0', index=True)
        create_time = db.Column(db.DateTime, unique=False, nullable=False, default=datetime.now)
        update_time = db.Column(db.DateTime, unique=False, nullable=False, default=datetime.now)

        pages = db.Column(db.Integer, unique=False, nullable=True)
        items = db.Column(db.Integer, unique=False, nullable=True)
        pid = db.Column(db.Integer, unique=False, nullable=True)  # Running
        start = db.Column(db.DateTime, unique=False, nullable=True, index=True)
        runtime = db.Column(db.String(20), unique=False, nullable=True)
        finish = db.Column(db.DateTime, unique=False, nullable=True, index=True)  # Finished
        href_log = db.Column(db.Text(), unique=False, nullable=True)
        href_items = db.Column(db.Text(), unique=False, nullable=True)

        def __repr__(self):
            return "<Job #%s in table %s, %s/%s/%s start: %s>" % (
                self.id, self.__tablename__, self.project, self.spider, self.job, self.start)

    return Job
    # sqlalchemy/ext/declarative/clsregistry.py:128: SAWarning: This declarative base already contains a class
    # with the same class name and module name as scrapydweb.models.Job,
    # and will be replaced in the string-lookup table.
    # https://stackoverflow.com/questions/27773489/dynamically-create-a-python-subclass-in-a-function
    # return type('Job_%s' % server, (Job, ), dict(__tablename__=server,  __bind_key__='jobs'))

# print(dir([create_table(s) for s in 'abc'][0]))


# http://flask-sqlalchemy.pocoo.org/2.3/models/    One-to-Many Relationships
# https://techarena51.com/blog/one-to-many-relationships-with-flask-sqlalchemy/
# https://docs.sqlalchemy.org/en/latest/orm/cascades.html#delete-orphan
# https://docs.sqlalchemy.org/en/latest/core/constraints.html#indexes
# https://stackoverflow.com/questions/14419299/adding-indexes-to-sqlalchemy-models-after-table-creation
# https://stackoverflow.com/questions/8890738/sqlalchemy-does-column-with-foreignkey-creates-index-automatically
class Task(db.Model):
    __tablename__ = 'task'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), unique=False, nullable=True)  # None
    trigger = db.Column(db.String(8), unique=False, nullable=False)  # cron, interval, date
    create_time = db.Column(db.DateTime, unique=False, nullable=False, default=datetime.now)  # datetime.utcnow
    update_time = db.Column(db.DateTime, unique=False, nullable=False, default=datetime.now)

    project = db.Column(db.String(255), unique=False, nullable=False)
    version = db.Column(db.String(255), unique=False, nullable=False)
    spider = db.Column(db.String(255), unique=False, nullable=False)
    jobid = db.Column(db.String(255), unique=False, nullable=False)
    settings_arguments = db.Column(db.Text(), unique=False, nullable=False)
    selected_nodes = db.Column(db.Text(), unique=False, nullable=False)

    year = db.Column(db.String(255), unique=False, nullable=False)
    month = db.Column(db.String(255), unique=False, nullable=False)
    day = db.Column(db.String(255), unique=False, nullable=False)
    week = db.Column(db.String(255), unique=False, nullable=False)
    day_of_week = db.Column(db.String(255), unique=False, nullable=False)
    hour = db.Column(db.String(255), unique=False, nullable=False)
    minute = db.Column(db.String(255), unique=False, nullable=False)
    second = db.Column(db.String(255), unique=False, nullable=False)

    start_date = db.Column(db.String(19), unique=False, nullable=True)  # '2019-01-01 00:00:01'     None
    end_date = db.Column(db.String(19), unique=False, nullable=True)  # '2019-01-01 00:00:01'       None

    timezone = db.Column(db.String(255), unique=False, nullable=True)  # None
    jitter = db.Column(db.Integer, unique=False, nullable=False)  # int
    misfire_grace_time = db.Column(db.Integer, unique=False, nullable=True)  # None|a positive integer
    coalesce = db.Column(db.String(5), unique=False, nullable=False)  # 'True'|'False'
    max_instances = db.Column(db.Integer, unique=False, nullable=False)  # int

    results = db.relationship('TaskResult', backref='task', cascade='all, delete-orphan', lazy=True)

    def __repr__(self):
        return "<Task #%s (%s), %s/%s/%s/%s, created at %s, updated at %s>" % (
                self.id, self.name, self.project, self.version, self.spider, self.jobid,
                self.create_time, self.update_time)


class TaskResult(db.Model):
    __tablename__ = 'task_result'

    id = db.Column(db.Integer, primary_key=True)
    task_id = db.Column(db.Integer, db.ForeignKey('task.id'), nullable=False, index=True)
    execute_time = db.Column(db.DateTime, unique=False, nullable=False, default=datetime.now)
    fail_count = db.Column(db.Integer, unique=False, nullable=False, default=0)
    pass_count = db.Column(db.Integer, unique=False, nullable=False, default=0)

    results = db.relationship('TaskJobResult', backref='task_result', cascade='all, delete-orphan', lazy=True)

    def __repr__(self):
        return "<TaskResult #%s of task #%s (%s), [FAIL %s, PASS %s], executed at %s>" % (
                self.id, self.task_id, self.task.name, self.fail_count, self.pass_count, self.execute_time)


class TaskJobResult(db.Model):
    __tablename__ = 'task_job_result'

    id = db.Column(db.Integer, primary_key=True)
    task_result_id = db.Column(db.Integer, db.ForeignKey('task_result.id'), nullable=False, index=True)
    run_time = db.Column(db.DateTime, unique=False, nullable=False, default=datetime.now)
    node = db.Column(db.Integer, unique=False, nullable=False, index=True)
    server = db.Column(db.String(255), unique=False, nullable=False)  # '127.0.0.1:6800'
    status_code = db.Column(db.Integer, unique=False, nullable=False)  # -1, 200
    status = db.Column(db.String(9), unique=False, nullable=False)  # ok|error|exception
    # psycopg2.DataError) value too long for type character varying(1000)
    # https://docs.sqlalchemy.org/en/latest/core/type_basics.html#sqlalchemy.types.Text
    # In general, TEXT objects do not have a length
    result = db.Column(db.Text(), unique=False, nullable=False)  # jobid|message|exception

    def __repr__(self):
        kwargs = dict(
            task_id=self.task_result.task_id,
            task_name=self.task_result.task.name,
            project=self.task_result.task.project,
            version=self.task_result.task.version,
            spider=self.task_result.task.spider,
            jobid=self.task_result.task.jobid,
            run_time=str(self.run_time),  # TypeError: Object of type datetime is not JSON serializable
            node=self.node,
            server=self.server,
            status_code=self.status_code,
            status=self.status,
            result=self.result,
            task_result_id=self.task_result_id,
            id=self.id,
        )
        return '<TaskJobResult \n%s>' % pformat(kwargs, indent=4)



# coding: utf-8
import re

from flask import render_template, request, url_for

from ...utils.send_email import send_email
from ..baseview import BaseView


class SendTextView(BaseView):

    def __init__(self):
        super(SendTextView, self).__init__()

        self.template = 'scrapydweb/send_text.html'

    def dispatch_request(self, **kwargs):
        kwargs = dict(
            node=self.node,
            url_slack=url_for('sendtextapi', opt='slack', channel_chatid_subject=None, text='some-text'),
            url_telegram=url_for('sendtextapi', opt='telegram', channel_chatid_subject=None, text='some-text'),
            url_email=url_for('sendtextapi', opt='email', channel_chatid_subject=None, text='some-text'),
        )
        return render_template(self.template, **kwargs)


class SendTextApiView(BaseView):
    # https://api.slack.com/methods/chat.postMessage
    # https://www.codementor.io/garethdwyer/building-a-telegram-bot-using-python-part-1-goi5fncay
    # slack_help = 'https://api.slack.com/apps'
    # telegram_help = 'https://core.telegram.org/bots#6-botfather'

    def __init__(self):
        super(SendTextApiView, self).__init__()

        self.opt = self.view_args['opt']
        self.opt = 'telegram' if self.opt == 'tg' else self.opt
        # https://stackoverflow.com/questions/10434599/get-the-data-received-in-a-flask-request
        # request.values: combined args and form, preferring args if keys overlap
        self.form = request.json or request.form

        if self.opt == 'email':
            self.channel_chatid_subject = (self.view_args['channel_chatid_subject']
                                           or request.args.get('subject', None)
                                           or self.form.get('subject', self.EMAIL_SUBJECT))
            # request.json['recipients'] could be a list type instead of a string type
            recipients = re.findall(r'[^\s"\',;\[\]]+@[^\s"\',;\[\]]+',
                                    request.args.get('recipients', '') or str(self.form.get('recipients', '')))
            self.EMAIL_KWARGS['email_recipients'] = recipients or self.EMAIL_RECIPIENTS
        elif self.opt == 'slack':
            self.channel_chatid_subject = (self.view_args['channel_chatid_subject']
                                           or request.args.get('channel', None)
                                           or self.form.get('channel', self.SLACK_CHANNEL))  # 'general'
        else:
            self.channel_chatid_subject = (self.view_args['channel_chatid_subject']
                                           or request.args.get('chat_id', None)
                                           or self.form.get('chat_id', self.TELEGRAM_CHAT_ID))
        self.logger.debug('channel_chatid_subject: %s', self.channel_chatid_subject)

        self.text = self.view_args['text'] or request.args.get('text', None)
        if not self.text:
            self.text = self.json_dumps(self.form) if self.form else 'test'
        self.logger.debug('text: %s', self.text)

        self.js = {}
        self.tested = False  # For test only

    def dispatch_request(self, **kwargs):
        if self.opt == 'email':
            self.send_email()
        elif self.opt == 'slack':
            self.send_slack()
        elif self.opt == 'telegram':
            self.send_telegram()
        self.js['when'] = self.get_now_string(True)
        return self.json_dumps(self.js, as_response=True)

    def send_email(self):
        if not self.EMAIL_PASSWORD:
            self.js = dict(status=self.ERROR, result="The EMAIL_PASSWORD option is unset")
            return
        self.EMAIL_KWARGS['subject'] = self.channel_chatid_subject
        self.EMAIL_KWARGS['content'] = self.text
        result, reason = send_email(to_retry=True, **self.EMAIL_KWARGS)
        if result is True:
            self.logger.debug("Sent to %s via Email", self.EMAIL_KWARGS['email_recipients'])
            self.js = dict(status=self.OK,
                           result=dict(reason=reason, sender=self.EMAIL_KWARGS['email_sender'],
                                       recipients=self.EMAIL_KWARGS['email_recipients'],
                                       subject=self.channel_chatid_subject, text=self.text))
        else:
            self.js = dict(status=self.ERROR, result=dict(reason=reason), debug=self.EMAIL_KWARGS)
            self.logger.error("Fail to send text via Email:\n%s", self.json_dumps(self.js))

    def send_slack(self):
        if not self.SLACK_TOKEN:
            self.js = dict(status=self.ERROR, result="The SLACK_TOKEN option is unset")
            return
        url = 'https://slack.com/api/chat.postMessage'
        data = dict(token=self.SLACK_TOKEN, channel=self.channel_chatid_subject, text=self.text)
        status_code, js = self.make_request(url, data=data, check_status=False)
        for key in ['auth', 'status', 'status_code', 'url', 'when']:
            js.pop(key, None)
        self.js = dict(url=url, status_code=status_code, result=js)
        # {"ok":false,"error":"invalid_auth"}
        # {"ok":false,"error":"channel_not_found"}
        # {"ok":false,"error":"no_text"}
        if js.get('ok', False):
            self.logger.debug("Sent to bot %s via Slack", js.get('message', {}).get('username', ''))
            self.js['status'] = self.OK
        else:
            self.js['status'] = self.ERROR
            if self.SLACK_TOKEN:
                self.js['debug'] = dict(token=self.SLACK_TOKEN, channel=self.channel_chatid_subject,
                                        text=self.text)
            self.logger.error("Fail to send text via Slack:\n%s", self.json_dumps(self.js))

    def send_telegram(self):
        if not self.TELEGRAM_TOKEN:
            self.js = dict(status=self.ERROR, result="The TELEGRAM_TOKEN option is unset")
            return
        url = 'https://api.telegram.org/bot%s/sendMessage' % self.TELEGRAM_TOKEN
        data = dict(text=self.text, chat_id=self.channel_chatid_subject)
        status_code, js = self.make_request(url, data=data, check_status=False)

        for key in ['auth', 'status', 'status_code', 'url', 'when']:
            js.pop(key, None)
        self.js = dict(url=url, status_code=status_code, result=js)
        if js.get('ok', False):
            self.logger.debug("Sent to %s via Telegram", js.get('result', {}).get('chat', {}).get('first_name', ''))
            self.js['status'] = self.OK
        # {"ok":false,"error_code":400,"description":"Bad Request: chat not found"}
        else:
            self.js['status'] = self.ERROR
            if self.TELEGRAM_TOKEN:
                self.js['debug'] = dict(token=self.TELEGRAM_TOKEN, chat_id=self.channel_chatid_subject,
                                        text=self.text)
            self.logger.error("Fail to send text via Telegram:\n%s", self.json_dumps(self.js))



# coding: utf-8
from tests.utils import cst, req


def multinode_command(app, client, opt, title, project, version_job=None):
    data = {'1': 'on', '2': 'on'}
    req(app, client, view='multinode', kws=dict(node=1, opt=opt, project=project, version_job=version_job),
        data=data, ins=[title, 'id="checkbox_1"', 'id="checkbox_2"'])


def test_multinode_stop(app, client):
    title = 'Stop Job (%s) of Project (%s)' % (cst.PROJECT, cst.JOBID)
    multinode_command(app, client, 'stop', title, cst.PROJECT, version_job=cst.JOBID)


def test_multinode_delproject(app, client):
    title = 'Delete Project (%s)' % cst.PROJECT
    multinode_command(app, client, 'delproject', title, cst.PROJECT)


def test_multinode_delversion(app, client):
    title = 'Delete Version (%s) of Project (%s)' % (cst.VERSION, cst.PROJECT)
    multinode_command(app, client, 'delversion', title, cst.PROJECT, version_job=cst.VERSION)



# -*- coding: utf-8 -*-

# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://doc.scrapy.org/en/latest/topics/item-pipeline.html


class ScrapydwebDemoPipeline(object):
    def process_item(self, item, spider):
        return item



# -*- coding: utf-8 -*-
import sys
import scrapy


class TestSpider(scrapy.Spider):
    name = 'test'
    allowed_domains = ['httpbin.org', 'google.com']
    # start_urls = ['http://httpbin.org/']

    custom_settings = dict(
        # JOBDIR='crawls/test',
        # ROBOTSTXT_OBEY=False,
        # USER_AGENT='Mozilla/5.0',
        # COOKIES_ENABLED=False,
        # CONCURRENT_REQUESTS=2,
        # DOWNLOAD_DELAY=1,
        # LOGSTATS_INTERVAL=1,
        FEED_EXPORT_ENCODING='utf-8'
    )

    num = 1

    def start_requests(self):
        # AttributeError: 'TestSpider' object has no attribute 'arg1'
        # self.logger.debug('self.arg1: %s' % self.arg1)  # self.arg1: val1
        # repr: 'Test\' "\xe6\xb5\x8b\xe8\xaf\x95'
        if sys.version_info[0] < 3 and getattr(self, 'arg1', None):
            self.logger.debug('self.arg1: ' + self.arg1.decode('utf-8'))
        else:
            self.logger.debug('self.arg1: %s' % getattr(self, 'arg1', None))  # self.arg1: None
        # self.logger.debug(self.settings.attributes.keys())
        # self.logger.debug(self.settings.attributes.values())
        # self.logger.debug(self.settings)
        # self.logger.debug(self.settings.attributes.items())
        if getattr(self, 'url', None):
            yield scrapy.Request(self.url)

        self.logger.debug('JOB: %s' % self.settings.get('JOB'))
        self.logger.debug('USER_AGENT: %s' % self.settings.get('USER_AGENT'))  # Scrapy/1.5.0 (+https://scrapy.org)
        self.logger.debug('ROBOTSTXT_OBEY: %s' % self.settings.getbool('ROBOTSTXT_OBEY'))  # True
        self.logger.debug('COOKIES_ENABLED: %s' % self.settings.getbool('COOKIES_ENABLED'))  # True
        self.logger.debug('CONCURRENT_REQUESTS: %s' % self.settings.getint('CONCURRENT_REQUESTS'))  # 16
        self.logger.debug('DOWNLOAD_DELAY: %s' % self.settings.getint('DOWNLOAD_DELAY'))  # 0
        self.logger.debug('CLOSESPIDER_TIMEOUT: %s' % self.settings.getint('CLOSESPIDER_TIMEOUT'))  # 0
        self.logger.debug('CLOSESPIDER_PAGECOUNT: %s' % self.settings.getint('CLOSESPIDER_PAGECOUNT'))  # 0

        self.log(u'Chinese characters: 汉字字符')
        self.logger.debug('2018-08-20 09:13:06 [apps_redis] DEBUG: Resuming crawl (675840 requests scheduled)')

        self.logger.warning('warn') #python2.7 AttributeError: 'LoggerAdapter' object has no attribute 'warn'
        self.logger.error('error')
        self.logger.warning('warning\n123abc')
        self.logger.error('error\n456abc')
        self.logger.error('error\n456abc')
        self.logger.critical('critical\n789abc')
        self.logger.warning('warning\n 123\nabc')
        self.logger.error('error\n 456\nabc')
        self.logger.critical('critical')
        self.logger.critical('critical\n 789\nabc')
        self.logger.critical('critical')
        self.logger.critical('critical')

        yield scrapy.Request('http://httpbin.org/redirect/1')
        yield scrapy.Request('http://httpbin.org/status/404')
        yield scrapy.Request('http://httpbin.org/headers')
        yield scrapy.Request('http://httpbin.org/headers')
        yield scrapy.Request('https://google.com/')

    def parse(self, response):
        self.log(response.text)
        if self.num == 1:
            yield scrapy.Request('https://www.baidu.com/')

        yield {u'Chinese 汉字 %s' % self.num: ''.join('0' + str(i) if i < 10 else str(i) for i in range(1, 100))}

        self.num += 1



# coding: utf-8
from flask import redirect, render_template, url_for

from ..baseview import BaseView


metadata = dict(
    project='',
    spider='',
    job='',
    selected_nodes=[]
)


class ClusterReportsView(BaseView):

    def __init__(self):
        super(ClusterReportsView, self).__init__()

        self.project = self.view_args['project'] or metadata['project']
        self.spider = self.view_args['spider'] or metadata['spider']
        self.job = self.view_args['job'] or metadata['job']
        self.selected_nodes = self.get_selected_nodes() or metadata['selected_nodes']
        metadata['project'] = self.project
        metadata['spider'] = self.spider
        metadata['job'] = self.job
        metadata['selected_nodes'] = self.selected_nodes

        self.template = 'scrapydweb/cluster_reports.html'

    def dispatch_request(self, **kwargs):
        if all([self.project, self.spider, self.job]):
            # Click reports memu for the second time
            if not any([self.view_args['project'], self.view_args['spider'], self.view_args['job']]):
                return redirect(url_for('clusterreports', node=self.node, project=self.project,
                                        spider=self.spider, job=self.job))
            # Click reports button on the Jobs page after reboot
            if not self.selected_nodes:
                return redirect(url_for('servers', node=self.node, opt='getreports', project=self.project,
                                        spider=self.spider, version_job=self.job))

        # Click reports memu for the first time
        if not any([self.project, self.spider, self.job]):
            url_servers = ''
        else:
            url_servers = url_for('servers', node=self.node, opt='getreports', project=self.project,
                                  spider=self.spider, version_job=self.job)

        kwargs = dict(
            node=self.node,
            project=self.project,
            spider=self.spider,
            job=self.job,
            selected_nodes=self.selected_nodes,
            url_report=url_for('log', node=self.node, opt='report', project=self.project,
                               spider=self.spider, job=self.job),
            url_servers=url_servers,
            url_jobs=url_for('jobs', node=self.node),
            # url_nodereports=url_for('nodereports', node=self.node),
        )
        return render_template(self.template, **kwargs)



# coding: utf-8
from functools import partial
import io
import os
import re

from flask import url_for

from scrapydweb.vars import DEMO_PROJECTS_PATH
from tests.utils import cst, req_single_scrapyd, set_single_scrapyd, upload_file_deploy


def test_auto_packaging_select_option(app, client):
    ins = [
        '(14 projects)',
        u"var folders = ['demo - 副本', 'demo',",
        "var projects = ['demo-copy', 'demo',",
        '<div>%s<' % cst.PROJECT,
        u'<div>demo - 副本<',
        '<div>demo<',
        '<div>demo_only_scrapy_cfg<'
    ]
    nos = ['<div>demo_without_scrapy_cfg<', '<h3>No projects found']
    req_single_scrapyd(app, client, view='deploy', kws=dict(node=1), ins=ins, nos=nos)

    for project in [cst.PROJECT, 'demo']:
        with io.open(os.path.join(cst.ROOT_DIR, 'data/%s/test' % project), 'w', encoding='utf-8') as f:
            f.write(u'')
        ins = ['id="folder_selected" value="%s"' % project, 'id="folder_selected_statement">%s<' % project]
        req_single_scrapyd(app, client, view='deploy', kws=dict(node=1), ins=ins)

    with io.open(os.path.join(cst.ROOT_DIR, 'data/demo/test'), 'w', encoding='utf-8') as f:
        f.write(u'')

    # SCRAPY_PROJECTS_DIR=os.path.join(cst.ROOT_DIR, 'data'),
    app.config['SCRAPY_PROJECTS_DIR'] = os.path.join(cst.ROOT_DIR, 'not-exist')
    req_single_scrapyd(app, client, view='deploy', kws=dict(node=1),
                       ins=['(0 projects)', '<h3>No projects found'])

    app.config['SCRAPY_PROJECTS_DIR'] = os.path.join(cst.ROOT_DIR, 'data', 'one_project_inside')
    req_single_scrapyd(app, client, view='deploy', kws=dict(node=1),
                       ins='(1 project)', nos='<h3>No projects found')

    if not os.environ.get('DATA_PATH', ''):
        app.config['SCRAPY_PROJECTS_DIR'] = ''
        req_single_scrapyd(app, client, view='deploy', kws=dict(node=1),
                           ins=DEMO_PROJECTS_PATH.replace('\\', '/'), nos='<h3>No projects found')


# {'status': 'error', 'message': 'Traceback
# ...TypeError:...activate_egg(eggpath)...\'tuple\' object is not an iterator\r\n'}
def test_addversion(app, client):
    data = {
        'project': 'fakeproject',
        'version': 'fakeversion',
        'file': (io.BytesIO(b'my file contents'), "fake.egg")
    }
    req_single_scrapyd(app, client, view='deploy.upload', kws=dict(node=1), data=data, ins='activate_egg')


# <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
# <title>Redirecting...</title>
# <h1>Redirecting...</h1>
# <p>You should be redirected automatically to target URL:
# <a href="/1/schedule/demo/2018-01-01T01_01_01/">/1/schedule/demo/2018-01-01T01_01_01/</a>.  If not click the link.
def test_auto_packaging(app, client):
    data = {
        'folder': cst.PROJECT,
        'project': cst.PROJECT,
        'version': cst.VERSION,
    }
    with app.test_request_context():
        # http://localhost/1/schedule/ScrapydWeb_demo/2018-01-01T01_01_01/
        req_single_scrapyd(app, client, view='deploy.upload', kws=dict(node=1), data=data,
                           location=url_for('schedule', node=1, project=cst.PROJECT, version=cst.VERSION))


def test_auto_packaging_unicode(app, client):
    if cst.WINDOWS_NOT_CP936:
        return
    data = {
        'folder': u'demo - 副本',
        'project': u'demo - 副本',
        'version': cst.VERSION,
    }
    with app.test_request_context():
        req_single_scrapyd(app, client, view='deploy.upload', kws=dict(node=1), data=data,
                           location=url_for('schedule', node=1, project='demo_____', version=cst.VERSION))


def test_scrapy_cfg(app, client):
    with app.test_request_context():
        for folder, result in cst.SCRAPY_CFG_DICT.items():
            data = {
                'folder': folder,
                'project': cst.PROJECT,
                'version': cst.VERSION,
            }
            if result:
                req_single_scrapyd(app, client, view='deploy.upload', kws=dict(node=1), data=data, ins=result)
            else:
                location = url_for('schedule', node=1, project=cst.PROJECT, version=cst.VERSION)
                req_single_scrapyd(app, client, view='deploy.upload', kws=dict(node=1), data=data, location=location)


def test_scrapy_cfg_node_not_exist(app, client):
    with app.test_request_context():
        for folder, result in cst.SCRAPY_CFG_DICT.items():
            data = {
                'folder': folder,
                'project': cst.PROJECT,
                'version': cst.VERSION,
            }
            nos = []
            if folder == 'demo_only_scrapy_cfg' or not result:
                ins = 'Fail to deploy project, got status'
            else:
                ins = ['Fail to deploy', result]
                nos = 'got status'
            req_single_scrapyd(app, client, view='deploy.upload', kws=dict(node=1), data=data,
                               ins=ins, nos=nos, set_to_second=True)


def test_upload_file_deploy(app, client):
    set_single_scrapyd(app)

    upload_file_deploy_singlenode = partial(upload_file_deploy, app=app, client=client, multinode=False)

    filenames = ['demo.egg', 'demo_inner.zip', 'demo_outer.zip',
                 'demo - Win7CNsendzipped.zip', 'demo - Win10cp1252.zip']
    if cst.WINDOWS_NOT_CP936:
        filenames.extend(['demo - Ubuntu.zip', 'demo - Ubuntu.tar.gz', 'demo - macOS.zip', 'demo - macOS.tar.gz'])
    else:
        filenames.extend([u'副本.zip', u'副本.tar.gz', u'副本.egg', u'demo - 副本 - Win7CN.zip',
                          u'demo - 副本 - Win7CNsendzipped.zip', u'demo - 副本 - Win10cp936.zip',
                          u'demo - 副本 - Ubuntu.zip', u'demo - 副本 - Ubuntu.tar.gz',
                          u'demo - 副本 - macOS.zip', u'demo - 副本 - macOS.tar.gz'])

    for filename in filenames:
        if filename == 'demo.egg':
            project = cst.PROJECT
            redirect_project = cst.PROJECT
        else:
            project = re.sub(r'\.egg|\.zip|\.tar\.gz', '', filename)
            project = 'demo_unicode' if project == u'副本' else project
            redirect_project = re.sub(cst.STRICT_NAME_PATTERN, '_', project)
        upload_file_deploy_singlenode(filename=filename, project=project, redirect_project=redirect_project)

    for filename, alert in cst.SCRAPY_CFG_DICT.items():
        if alert:
            upload_file_deploy_singlenode(filename='%s.zip' % filename, project=filename, alert=alert, fail=True)
        else:
            upload_file_deploy_singlenode(filename='%s.zip' % filename, project=filename, redirect_project=filename)

    app.config['SCRAPYD_SERVERS'] = ['not-exist:6801']

    for filename, alert in cst.SCRAPY_CFG_DICT.items():
        if filename == 'demo_only_scrapy_cfg' or not alert:
            alert = 'Fail to deploy project, got status'
        upload_file_deploy_singlenode(filename='%s.zip' % filename, project=filename, alert=alert, fail=True)



# coding: utf-8
from flask import render_template, url_for

from ..baseview import BaseView


class MultinodeView(BaseView):
    methods = ['POST']

    def __init__(self):
        super(MultinodeView, self).__init__()

        self.opt = self.view_args['opt']
        self.project = self.view_args['project']
        self.version_job = self.view_args['version_job']

        self.template = 'scrapydweb/multinode_results.html'

    def dispatch_request(self, **kwargs):
        selected_nodes = self.get_selected_nodes()
        url_xhr = url_for('api', node=selected_nodes[0], opt=self.opt,
                          project=self.project, version_spider_job=self.version_job)

        if self.opt == 'stop':
            title = "Stop Job (%s) of Project (%s)" % (self.project, self.version_job)
            url_servers = url_for('servers', node=self.node, opt='listjobs', project=self.project)
            btn_servers = "Servers &raquo; List Running Jobs"
        elif self.opt == 'delversion':
            title = "Delete Version (%s) of Project (%s)" % (self.version_job, self.project)
            url_servers = url_for('servers', node=self.node, opt='listversions', project=self.project)
            btn_servers = "Servers &raquo; List Versions"
        else:  # elif opt == 'delproject':
            title = "Delete Project (%s)" % self.project
            url_servers = url_for('servers', node=self.node, opt='listprojects', project=self.project)
            btn_servers = "Servers &raquo; List Projects"

        kwargs = dict(
            node=self.node,
            title=title,
            opt=self.opt,
            project=self.project,
            version_job=self.version_job,
            selected_nodes=selected_nodes,
            url_xhr=url_xhr,
            url_servers=url_servers,
            btn_servers=btn_servers,
            url_projects_list=[url_for('projects', node=n) for n in range(1, self.SCRAPYD_SERVERS_AMOUNT + 1)]
        )
        return render_template(self.template, **kwargs)



# coding: utf-8
from collections import OrderedDict
from datetime import datetime
import re
import traceback

from flask import flash, get_flashed_messages, render_template, request, url_for
from six.moves.urllib.parse import urljoin

from ...common import handle_metadata
from ...models import create_jobs_table, db
from ...vars import STRICT_NAME_PATTERN, jobs_table_map
from ..baseview import BaseView


_metadata = handle_metadata()
metadata = dict(
    pageview=_metadata.get('pageview', 1),
    per_page=_metadata.get('jobs_per_page', 100),
    style=_metadata.get('jobs_style', 'database'),
    unique_key_strings={}
)

STATUS_PENDING = '0'
STATUS_RUNNING = '1'
STATUS_FINISHED = '2'
NOT_DELETED = '0'
DELETED = '1'
HREF_PATTERN = re.compile(r"""href=['"](.+?)['"]""")  # Temp support for Scrapyd v1.3.0 (not released)
JOB_PATTERN = re.compile(r"""
                            <tr>
                                <td>(?P<Project>.*?)</td>
                                <td>(?P<Spider>.*?)</td>
                                <td>(?P<Job>.*?)</td>
                                (?:<td>(?P<PID>.*?)</td>)?
                                (?:<td>(?P<Start>.*?)</td>)?
                                (?:<td>(?P<Runtime>.*?)</td>)?
                                (?:<td>(?P<Finish>.*?)</td>)?
                                (?:<td>(?P<Log>.*?)</td>)?
                                (?:<td>(?P<Items>.*?)</td>)?
                                [\w\W]*?  # Temp support for Scrapyd v1.3.0 (not released)
                            </tr>
                          """, re.X)
JOB_KEYS = ['project', 'spider', 'job', 'pid', 'start', 'runtime', 'finish', 'href_log', 'href_items']


class JobsView(BaseView):
    # methods = ['GET']
    metadata = metadata

    def __init__(self):
        super(JobsView, self).__init__()

        style = request.args.get('style')
        self.style = style if style in ['database', 'classic'] else self.metadata['style']
        if self.style != self.metadata['style']:
            self.metadata['style'] = self.style
            handle_metadata('jobs_style', self.style)
            self.logger.debug("Change style to %s", self.metadata['style'])

        self.per_page = request.args.get('per_page', default=self.metadata['per_page'], type=int)
        if self.per_page != self.metadata['per_page']:
            self.metadata['per_page'] = self.per_page
            handle_metadata('jobs_per_page', self.per_page)
            self.logger.debug("Change per_page to %s", self.metadata['per_page'])
        self.page = request.args.get('page', default=1, type=int)

        self.url = 'http://%s/jobs' % self.SCRAPYD_SERVER
        if self.SCRAPYD_SERVER_PUBLIC_URL:
            self.public_url = '%s/jobs' % self.SCRAPYD_SERVER_PUBLIC_URL
        else:
            self.public_url = ''
        self.text = ''
        self.kwargs = {}
        if self.USE_MOBILEUI:
            self.style = 'classic'
            self.template = 'scrapydweb/jobs_mobileui.html'
        elif self.style == 'classic':
            self.template = 'scrapydweb/jobs_classic.html'
        else:  # 'database'
            self.template = 'scrapydweb/jobs.html'

        self.listjobs = request.args.get('listjobs', None)

        self.liststats_datas = {}
        self.jobs_dict = {}

        self.jobs = []
        self.jobs_backup = []
        self.pending_jobs = []
        self.running_jobs = []
        self.finished_jobs = []
        self.jobs_pagination = None

        self.Job = None  # database class Job

    def dispatch_request(self, **kwargs):
        status_code, self.text = self.make_request(self.url, auth=self.AUTH, as_json=False)
        if status_code != 200 or not re.search(r'<body><h1>Jobs</h1>', self.text):
            kwargs = dict(
                node=self.node,
                url=self.url,
                status_code=status_code,
                text=self.text,
                tip="Click the above link to make sure your Scrapyd server is accessable. "
            )
            return render_template(self.template_fail, **kwargs)
        # Temp support for Scrapyd v1.3.0 (not released)
        self.text = re.sub(r'<thead>.*?</thead>', '', self.text, flags=re.S)
        self.jobs = [dict(zip(JOB_KEYS, job)) for job in re.findall(JOB_PATTERN, self.text)]
        self.jobs_backup = list(self.jobs)

        if self.listjobs:
            return self.json_dumps(self.jobs, as_response=True)

        if self.POST:  # To update self.liststats_datas
            self.get_liststats_datas()
        else:
            self.metadata['pageview'] += 1
            self.logger.debug('metadata: %s', self.metadata)
            self.set_flash()
        if self.style == 'database' or self.POST:
            self.handle_jobs_with_db()
        if self.POST:
            try:
                self.set_jobs_dict()
            except:
                raise
            finally:
                get_flashed_messages()
            return self.json_dumps(self.jobs_dict, as_response=True)
        if self.style != 'database':
            self.jobs = self.jobs_backup
            self.handle_jobs_without_db()
        self.set_kwargs()
        return render_template(self.template, **self.kwargs)

    def set_flash(self):
        if self.metadata['pageview'] > 2 and self.metadata['pageview'] % 100:
            return
        if not self.ENABLE_AUTH and self.SCRAPYD_SERVERS_AMOUNT == 1:
            flash("Set 'ENABLE_AUTH = True' to enable basic auth for web UI", self.INFO)
        if self.IS_LOCAL_SCRAPYD_SERVER:
            if not self.LOCAL_SCRAPYD_LOGS_DIR:
                flash(("Set up the LOCAL_SCRAPYD_LOGS_DIR option to speed up the loading of scrapy logfiles "
                      "for the LOCAL_SCRAPYD_SERVER %s" % self.SCRAPYD_SERVER), self.WARN)
            if not self.ENABLE_LOGPARSER:
                flash("Set 'ENABLE_LOGPARSER = True' to run LogParser as a subprocess at startup", self.WARN)
        if not self.ENABLE_MONITOR and self.SCRAPYD_SERVERS_AMOUNT == 1:
            flash("Set 'ENABLE_MONITOR = True' to enable the monitor feature", self.INFO)

# stats.json by LogParser
# {
#     "status_code": 200,
#     "status": "ok",
#     "datas": {
#         "demo": {
#             "test": {
#                 "2019-01-01T0_00_01": {
#                     "pages": 3,
#                     "items": 2,
    def get_liststats_datas(self):
        # NOTE: get_response_from_view() would update g.url_jobs_list, unexpected for mobileui
        # request.url: http://localhost/1/api/liststats/
        # TODO: test https
        url_liststats = url_for('api', node=self.node, opt='liststats')
        js = self.get_response_from_view(url_liststats, as_json=True)
        if js['status'] == self.OK:
            self.liststats_datas = js.pop('datas', {})
            self.logger.debug("Got datas with %s entries from liststats: %s", len(self.liststats_datas), js)
        else:
            self.logger.warning("Fail to get datas from liststats: (%s) %s %s",
                                js['status_code'], js['status'], js.get('tip', ''))

    def create_table(self):
        self.Job = jobs_table_map.get(self.node, None)
        if self.Job is not None:
            self.logger.debug("Got table: %s", self.Job.__tablename__)
        else:
            self.Job = create_jobs_table(re.sub(STRICT_NAME_PATTERN, '_', self.SCRAPYD_SERVER))
            # sqlite3.OperationalError: table "127_0_0_1_6800" already exists
            db.create_all(bind='jobs')
            self.metadata[self.node] = self.Job
            jobs_table_map[self.node] = self.Job
            self.logger.debug("Created table: %s", self.Job.__tablename__)

    def handle_jobs_with_db(self):
        try:
            if request.args.get('raise_exception') == 'True':  # For test only
                assert False, "raise_exception: True"
            self.handle_unique_constraint()
            self.create_table()
            self.db_insert_jobs()
            self.db_clean_pending_jobs()
            self.query_jobs()
        except Exception as err:
            self.logger.error("Fail to persist jobs in database: %s", traceback.format_exc())
            db.session.rollback()
            flash("Fail to persist jobs in database: %s" % err, self.WARN)
            # sqlalchemy.exc.InvalidRequestError: Table '127_0_0_1_6800' is already defined for this MetaData instance.
            # Specify 'extend_existing=True' to redefine options and columns on an existing Table object.
            if "is already defined for this MetaData instance" in str(err):
                flash("Please restart ScrapydWeb to work around this occasional bug!", self.WARN)
            if self.style == 'database' and not self.POST:
                self.style = 'classic'
                self.template = 'scrapydweb/jobs_classic.html'
                self.metadata['style'] = self.style
                handle_metadata('jobs_style', self.style)
                msg = "Change style to %s" % self.style
                self.logger.info(msg)
                # flash(msg, self.WARN)

    # Note that there may be jobs with the same combination of (project, spider, job) in the fetched Jobs
    def handle_unique_constraint(self):
        seen_jobs = OrderedDict()
        for job in self.jobs:  # (Pending, Running) ASC
            if job['finish']:
                break
            unique_key = (job['project'], job['spider'], job['job'])
            if unique_key in seen_jobs:  # ignore previous
                start = seen_jobs[unique_key]['start']
                finish = seen_jobs[unique_key]['finish']
                unique_key_string = '/'.join(list(unique_key) + [start, finish, str(self.node)])
                if start:
                    msg = "Ignore seen running job: %s, started at %s" % ('/'.join(unique_key), start)
                else:
                    msg = "Ignore seen pending job: %s" % ('/'.join(unique_key))
                self.logger.debug(msg)
                if unique_key_string not in self.metadata['unique_key_strings']:  # flash only once
                    self.metadata['unique_key_strings'][unique_key_string] = None
                    flash(msg, self.WARN if start else self.INFO)
                seen_jobs.pop(unique_key)
            seen_jobs[unique_key] = job
        for job in reversed(self.jobs):  # Finished DESC
            if not job['finish']:
                break
            unique_key = (job['project'], job['spider'], job['job'])
            if unique_key in seen_jobs:  # ignore current
                unique_key_string = '/'.join(list(unique_key) + [job['start'], job['finish'], str(self.node)])
                msg = "Ignore seen finished job: %s, started at %s" % ('/'.join(unique_key), job['start'])
                self.logger.debug(msg)
                if unique_key_string not in self.metadata['unique_key_strings']:
                    self.metadata['unique_key_strings'][unique_key_string] = None
                    flash(msg, self.INFO)
            else:
                seen_jobs[unique_key] = job
        self.jobs = list(seen_jobs.values())

    def db_insert_jobs(self):
        records = []
        for job in self.jobs:  # set(self.jobs): unhashable type: 'dict'
            record = self.Job.query.filter_by(project=job['project'], spider=job['spider'], job=job['job']).first()
            if record:
                self.logger.debug("Found job in database: %s", record)
                if record.deleted == DELETED:
                    if record.status == STATUS_FINISHED and str(record.start) == job['start']:
                        self.logger.info("Ignore deleted job: %s", record)
                        continue
                    else:
                        record.deleted = NOT_DELETED
                        record.pages = None
                        record.items = None
                        self.logger.info("Recover deleted job: %s", record)
                        flash("Recover deleted job: %s" % job, self.WARN)
            else:
                record = self.Job()
            records.append(record)
            for k, v in job.items():
                v = v or None  # Save NULL in database for empty string
                if k in ['start', 'finish']:
                    v = datetime.strptime(v, '%Y-%m-%d %H:%M:%S') if v else None  # Avoid empty string
                elif k in ['href_log', 'href_items']:  # <a href='/logs/demo/test/xxx.log'>Log</a>
                    m = re.search(HREF_PATTERN, v) if v else None
                    v = m.group(1) if m else v
                setattr(record, k, v)
            if not job['start']:
                record.status = STATUS_PENDING
            elif not job['finish']:
                record.status = STATUS_RUNNING
            else:
                record.status = STATUS_FINISHED
            if not job['start']:
                record.pages = None
                record.items = None
            elif self.liststats_datas:
                try:
                    data = self.liststats_datas[job['project']][job['spider']][job['job']]
                    record.pages = data['pages']  # Logparser: None or non-negative int
                    record.items = data['items']  # Logparser: None or non-negative int
                except KeyError:
                    pass
                except Exception as err:
                    self.logger.error(err)
            # SQLite DateTime type only accepts Python datetime and date objects as input
            record.update_time = datetime.now()  # datetime.now().replace(microsecond=0)
        # https://www.reddit.com/r/flask/comments/3tea4k/af_flasksqlalchemy_bulk_updateinsert/
        db.session.add_all(records)
        db.session.commit()

    def db_clean_pending_jobs(self):
        current_pending_jobs = [(job['project'], job['spider'], job['job'])
                                for job in self.jobs_backup if not job['start']]
        for record in self.Job.query.filter_by(start=None).all():
            if (record.project, record.spider, record.job) not in current_pending_jobs:
                db.session.delete(record)
                db.session.commit()
                self.logger.info("Deleted pending jobs %s", record)

    def query_jobs(self):
        current_running_job_pids = [int(job['pid']) for job in self.jobs_backup if job['pid']]
        self.logger.debug("current_running_job_pids: %s", current_running_job_pids)
        self.jobs_pagination = self.Job.query.filter_by(deleted=NOT_DELETED).order_by(
            self.Job.status.asc(), self.Job.finish.desc(), self.Job.start.asc(), self.Job.id.asc()).paginate(
            page=self.page, per_page=self.per_page, error_out=False)
        with db.session.no_autoflush:
            for index, job in enumerate(self.jobs_pagination.items,
                                        (self.jobs_pagination.page - 1) * self.jobs_pagination.per_page + 1):
                # print(vars(job))
                job.index = index
                job.pid = job.pid or ''
                job.start = job.start or ''  # None for Pending jobs
                job.runtime = job.runtime or ''
                job.finish = job.finish or ''  # None for Pending and Running jobs
                job.update_time = self.remove_microsecond(job.update_time)
                job.to_be_killed = True if job.pid and job.pid not in current_running_job_pids else False
                if job.finish:
                    job.url_multinode = url_for('servers', node=self.node, opt='schedule', project=job.project,
                                                version_job=self.DEFAULT_LATEST_VERSION, spider=job.spider)
                    job.url_action = url_for('schedule', node=self.node, project=job.project,
                                             version=self.DEFAULT_LATEST_VERSION, spider=job.spider)
                else:
                    job.url_multinode = url_for('servers', node=self.node, opt='stop', project=job.project,
                                                version_job=job.job)
                    job.url_action = url_for('api', node=self.node, opt='stop', project=job.project,
                                             version_spider_job=job.job)
                if job.start:
                    job.pages = self.NA if job.pages is None else job.pages  # May be 0
                    job.items = self.NA if job.items is None else job.items  # May be 0
                else:  # Pending
                    job.pages = None  # from Running/Finished to Pending
                    job.items = None
                    continue
                job_finished = 'True' if job.finish else None
                job.url_utf8 = url_for('log', node=self.node, opt='utf8', project=job.project, ui=self.UI,
                                       spider=job.spider, job=job.job, job_finished=job_finished)
                job.url_stats = url_for('log', node=self.node, opt='stats', project=job.project, ui=self.UI,
                                        spider=job.spider, job=job.job, job_finished=job_finished)
                job.url_clusterreports = url_for('clusterreports', node=self.node, project=job.project,
                                                 spider=job.spider, job=job.job)
                # '/items/demo/test/2018-10-12_205507.log'
                job.url_source = urljoin(self.public_url or self.url, job.href_log)
                if job.href_items:
                    job.url_items = urljoin(self.public_url or self.url, job.href_items)
                else:
                    job.url_items = ''
                job.url_delete = url_for('jobs.xhr', node=self.node, action='delete', id=job.id)

    def set_jobs_dict(self):
        for job in self.jobs_pagination.items:  # Pagination obj in handle_jobs_with_db() > query_jobs()
            key = '%s/%s/%s' % (job.project, job.spider, job.job)
            value = dict((k, v) for (k, v) in job.__dict__.items() if not k.startswith('_'))
            for k, v in value.items():
                if k in ['create_time', 'update_time', 'start', 'finish']:
                    value[k] = str(value[k])
            self.jobs_dict[key] = value

    def handle_jobs_without_db(self):
        for job in self.jobs:
            job['start'] = job['start'][5:]
            job['finish'] = job['finish'][5:]
            if not job['start']:
                self.pending_jobs.append(job)
            else:
                if job['finish']:
                    self.finished_jobs.append(job)
                    job['url_multinode_run'] = url_for('servers', node=self.node, opt='schedule',
                                                       project=job['project'], version_job=self.DEFAULT_LATEST_VERSION,
                                                       spider=job['spider'])
                    job['url_schedule'] = url_for('schedule', node=self.node, project=job['project'],
                                                  version=self.DEFAULT_LATEST_VERSION, spider=job['spider'])
                    job['url_start'] = url_for('api', node=self.node, opt='start', project=job['project'],
                                               version_spider_job=job['spider'])
                else:
                    self.running_jobs.append(job)
                    job['url_forcestop'] = url_for('api', node=self.node, opt='forcestop', project=job['project'],
                                                   version_spider_job=job['job'])

                job_finished = 'True' if job['finish'] else None
                job['url_utf8'] = url_for('log', node=self.node, opt='utf8', project=job['project'], ui=self.UI,
                                          spider=job['spider'], job=job['job'], job_finished=job_finished)
                job['url_stats'] = url_for('log', node=self.node, opt='stats', project=job['project'], ui=self.UI,
                                           spider=job['spider'], job=job['job'], job_finished=job_finished)
                job['url_clusterreports'] = url_for('clusterreports', node=self.node, project=job['project'],
                                                    spider=job['spider'], job=job['job'])
                # <a href='/items/demo/test/2018-10-12_205507.jl'>Items</a>
                m = re.search(HREF_PATTERN, job['href_items'])
                if m:
                    job['url_items'] = urljoin(self.public_url or self.url, m.group(1))
                else:
                    job['url_items'] = ''

            if not job['finish']:
                job['url_multinode_stop'] = url_for('servers', node=self.node, opt='stop', project=job['project'],
                                                    version_job=job['job'])
                job['url_stop'] = url_for('api', node=self.node, opt='stop', project=job['project'],
                                          version_spider_job=job['job'])

    def set_kwargs(self):
        self.kwargs = dict(
            node=self.node,
            url=self.url,
            url_schedule=url_for('schedule', node=self.node),
            url_liststats=url_for('api', node=self.node, opt='liststats'),
            url_liststats_source='http://%s/logs/stats.json' % self.SCRAPYD_SERVER,
            SCRAPYD_SERVER=self.SCRAPYD_SERVER.split(':')[0],
            LOGPARSER_VERSION=self.LOGPARSER_VERSION,
            JOBS_RELOAD_INTERVAL=self.JOBS_RELOAD_INTERVAL,
            IS_IE_EDGE=self.IS_IE_EDGE,
            pageview=self.metadata['pageview'],
            FEATURES=self.FEATURES
        )
        if self.style == 'database':
            self.kwargs.update(dict(
                url_jobs_classic=url_for('jobs', node=self.node, style='classic'),
                jobs=self.jobs_pagination
            ))
            return

        if self.JOBS_FINISHED_JOBS_LIMIT > 0:
            self.finished_jobs = self.finished_jobs[::-1][:self.JOBS_FINISHED_JOBS_LIMIT]
        else:
            self.finished_jobs = self.finished_jobs[::-1]
        self.kwargs.update(dict(
            colspan=14,
            url_jobs_database=url_for('jobs', node=self.node, style='database'),
            pending_jobs=self.pending_jobs,
            running_jobs=self.running_jobs,
            finished_jobs=self.finished_jobs,
            SHOW_JOBS_JOB_COLUMN=self.SHOW_JOBS_JOB_COLUMN
        ))


class JobsXhrView(BaseView):
    metadata = metadata

    def __init__(self):
        super(JobsXhrView, self).__init__()

        self.action = self.view_args['action']  # delete
        self.id = self.view_args['id']  # <int:id>

        self.js = {}
        self.Job = jobs_table_map[self.node]  # database class Job

    def dispatch_request(self, **kwargs):
        job = self.Job.query.get(self.id)
        if job:
            try:
                job.deleted = DELETED
                db.session.commit()
            except Exception as err:
                self.logger.error(traceback.format_exc())
                db.session.rollback()
                self.js['status'] = self.ERROR
                self.js['message'] = str(err)
            else:
                self.js['status'] = self.OK
                self.logger.info(self.js.setdefault('tip', "Deleted %s" % job))
        else:
            self.js['status'] = self.ERROR
            self.js['message'] = "job #%s not found in the database" % self.id

        return self.json_dumps(self.js, as_response=True)



# coding: utf-8
from collections import OrderedDict
import errno
import os


class Slot:
    def __init__(self, limit_egg=10, limit_data=10):
        self.limit_egg = limit_egg
        self.limit_data = limit_data
        self._egg = OrderedDict()
        self._data = OrderedDict()

    @property
    def egg(self):
        return self._egg

    @property
    def data(self):
        return self._data

    def add_egg(self, key, value):
        self._egg[key] = value
        if len(self._egg) > self.limit_egg:
            self._egg.popitem(last=False)

    def add_data(self, key, value):
        self._data[key] = value
        if len(self._data) > self.limit_data:
            self._data.popitem(last=False)


slot = Slot()


# https://stackoverflow.com/a/600612/10517783
def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as err:  # Python >2.5
        if err.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



# coding: utf-8
from flask import url_for

from tests.utils import cst, req_single_scrapyd


# Location: http://127.0.0.1:5000/1/jobs/?ui=mobile
def test_index(app, client):
    with app.test_request_context():
        for __, headers in cst.HEADERS_DICT.items():
            req_single_scrapyd(app, client, view='index', kws=dict(ui='mobile'), headers=headers,
                               location=url_for('jobs', node=1, ui='mobile'))

        for key in ['Chrome', 'iPad']:
            req_single_scrapyd(app, client, view='index', kws={}, headers=cst.HEADERS_DICT[key],
                               location=url_for('jobs', node=1))  # not the Servers page

        for key in ['iPhone', 'Android']:
            req_single_scrapyd(app, client, view='index', kws={}, headers=cst.HEADERS_DICT[key],
                               location=url_for('jobs', node=1, ui='mobile'))


def test_check_browser(app, client):
    ins = 'checkBrowser();'
    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1), headers=cst.HEADERS_DICT['IE'], ins=ins)
    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1), headers=cst.HEADERS_DICT['EDGE'], ins=ins)


def test_dropdown_for_mobile_device(app, client):
    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1), headers=cst.HEADERS_DICT['Chrome'],
                       ins='dropdown.css', nos=['dropdown_mobileui.css', 'handleDropdown();'])
    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1), headers=cst.HEADERS_DICT['iPhone'],
                       nos='dropdown.css', ins=['dropdown_mobileui.css', 'handleDropdown();'])
    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1), headers=cst.HEADERS_DICT['iPad'],
                       nos='dropdown.css', ins=['dropdown_mobileui.css', 'handleDropdown();'])


def test_check_update(app, client):
    @app.context_processor
    def inject_variable():
        return dict(CHECK_LATEST_VERSION_FREQ=1)

    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1),
                       ins='<script>setTimeout("checkLatestVersion(',
                       nos='<!-- <script>setTimeout("checkLatestVersion(')

    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1, ui='mobile'), mobileui=True,
                       ins='<script>setTimeout("checkLatestVersion(',
                       nos='<!-- <script>setTimeout("checkLatestVersion(')

    @app.context_processor
    def inject_variable():
        return dict(CHECK_LATEST_VERSION_FREQ=100)

    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1), nos='<script>setTimeout("checkLatestVersion(')
    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1, ui='mobile'), mobileui=True,
                       nos='<script>setTimeout("checkLatestVersion(')


def test_page(app, client):
    for view, title in cst.VIEW_TITLE_MAP.items():
        req_single_scrapyd(app, client, view=view, kws=dict(node=1), ins=title)


def test_select_multinode_checkbox(app, client):
    for view in ['deploy', 'schedule']:
        req_single_scrapyd(app, client, view=view, kws=dict(node=1), nos='CheckAll / UncheckAll')


def test_items(app, client):
    try:
        req_single_scrapyd(app, client, view='items', kws=dict(node=1), ins='Directory listing for /items/')
    except AssertionError:
        req_single_scrapyd(app, client, view='items', kws=dict(node=1), ins='No Such Resource')


def test_switch_node_skip(app, client):
    req_single_scrapyd(app, client, view='jobs', kws=dict(node=1),
                       nos=['onclick="switchNode', 'id="skip_nodes_checkbox"'])


# <span>Cluster Reports</span>
# <el-tab-pane label="Get Reports" name="getreports">
def test_cluster_reports_not_exists(app, client):
    nos = ['<span>Cluster Reports</span>', '<el-tab-pane label="Get Reports"']
    req_single_scrapyd(app, client, view='servers', kws=dict(node=1), nos=nos)



# coding: utf-8
import io
import os
import re

from setuptools import find_packages, setup


CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

about = {}
with open(os.path.join(CURRENT_DIR, 'scrapydweb', '__version__.py')) as f:
    exec(f.read(), about)

with io.open("README.md", 'r', encoding='utf-8') as f:
    long_description = re.sub(r':\w+:\s', '', f.read())  # Remove emoji


setup(
    name=about['__title__'],
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    license=about['__license__'],
    description=about['__description__'],

    long_description=long_description,
    long_description_content_type="text/markdown",

    packages=find_packages(exclude=("tests", )),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    install_requires=[
        "APScheduler==3.6.0",  # Mar 18, 2019
        "click==7.1.2",  # Apr 28, 2020
        "colorama==0.4.0",  # Oct 10, 2018
        "Flask==2.0.0",  # May 12, 2021
        "Flask-Compress==1.4.0",  # Jan 5, 2017
        "Flask-SQLAlchemy==2.4.0",  # Apr 25, 2019
        "idna==2.7",  # Jun 11, 2018
        "itsdangerous==2.0.0",  # May 12, 2021
        "Jinja2==3.0.0",  # May 12, 2021
        "logparser==0.8.2",
        "MarkupSafe==2.0.0",  # May 12, 2021
        "pexpect==4.7.0",  # Apr 7, 2019
        "ptyprocess==0.6.0",  # Jun 22, 2018
        "pytz==2018.9",  # Jan 7, 2019
        "requests>=2.21.0",  # Dec 10, 2018
        "setuptools>=40.6.3",  # Dec 11, 2018
        "six==1.16.0",  # May 5, 2021
        "SQLAlchemy==1.3.24",  # Mar 31, 2021
        "tzlocal==1.5.1",  # Dec 1, 2017
        "w3lib==2.0.0",  # Aug 11, 2022
        "Werkzeug==2.0.0",  # May 12, 2021
    ],

    entry_points={
        "console_scripts": {
            "scrapydwebx = scrapydweb.run:main"
        }
    },

    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ]
)



# coding: utf-8
from datetime import datetime
import json
import logging
import traceback

from flask import Blueprint, flash, render_template, request, send_file, url_for

from ...common import handle_metadata
from ...models import Task, TaskResult, TaskJobResult, db
from ...vars import SCHEDULER_STATE_DICT, STATE_PAUSED, STATE_RUNNING, TIMER_TASKS_HISTORY_LOG
from ..baseview import BaseView


apscheduler_logger = logging.getLogger('apscheduler')
metadata = dict(per_page=handle_metadata().get('tasks_per_page', 100))

bp = Blueprint('tasks', __name__, url_prefix='/')


@bp.route('/tasks/history/')
def history():
    return send_file(TIMER_TASKS_HISTORY_LOG, mimetype='text/plain', cache_timeout=0)


# https://apscheduler.readthedocs.io/en/latest/userguide.html
# https://apscheduler.readthedocs.io/en/latest/modules/schedulers/base.html#module-apscheduler.schedulers.base
# https://apscheduler.readthedocs.io/en/latest/modules/job.html#apscheduler.job.Job
class TasksView(BaseView):
    metadata = metadata

    def __init__(self):
        super(TasksView, self).__init__()

        self.task_id = self.view_args['task_id']  # <int:task_id>, 0 ok, -1 fail
        self.task_result_id = self.view_args['task_result_id']  # <int:task_result_id>

        self.flash = request.args.get('flash', '')
        self.per_page = request.args.get('per_page', default=self.metadata['per_page'], type=int)
        if self.per_page != self.metadata['per_page']:
            self.metadata['per_page'] = self.per_page
            handle_metadata('tasks_per_page', self.per_page)
            self.logger.debug("Change per_page to %s", self.metadata['per_page'])
        self.page = request.args.get('page', default=1, type=int)

        # If self.task is defined before handle_metadata('tasks_per_page', self.per_page)
        # Instance <Task at 0x58140f0> is not bound to a Session;
        # attribute refresh operation cannot proceed
        # (Background on this error at: http://sqlalche.me/e/bhk3)
        self.task = Task.query.get(self.task_id) if self.task_id else None
        self.kwargs = {}
        self.template = ''

    def dispatch_request(self, **kwargs):
        if self.flash:
            flash(self.flash, self.INFO)

        # http://flask-sqlalchemy.pocoo.org/2.3/queries/#queries-in-views
        # Use get_or_404 to ensure that url_for the Stats page works
        # task = Task.query.get_or_404(self.task_id)
        if self.task_id and not self.task:
            message = "Task #%s not found" % self.task_id
            self.logger.error(message)
            return render_template(self.template_fail, node=self.node, message=message)

        if self.task_id and self.task_result_id:
            self.template = 'scrapydweb/task_job_results.html'
            self.query_task_job_results()
        elif self.task_id:
            self.template = 'scrapydweb/task_results.html'
            self.query_task_results()
        else:
            self.template = 'scrapydweb/tasks.html'
            self.query_tasks()
        return render_template(self.template, **self.kwargs)

    def query_tasks(self):
        if self.scheduler.state == STATE_PAUSED:
            flash("Click the DISABLED button to enable the scheduler for timer tasks. ", self.WARN)
        self.remove_apscheduler_job_without_task()

        # https://stackoverflow.com/questions/43103585/python-flask-sqlalchemy-pagination
        # https://blog.miguelgrinberg.com/post/the-flask-mega-tutorial-part-ix-pagination
        # http://flask-sqlalchemy.pocoo.org/2.3/api/#flask_sqlalchemy.BaseQuery.paginate
        # paginate(page=None, per_page=None, error_out=True, max_per_page=None)
        # tasks = Task.query.all()
        tasks = Task.query.order_by(Task.id.desc()).paginate(
            page=self.page, per_page=self.per_page, error_out=False)
        self.process_tasks(tasks)

        # default-sort in Vue would cause background color of the buttons flash once
        # :default-sort="{prop: 'status', order: 'descending'}"
        tasks.items.sort(key=lambda task: task.status, reverse=True)

        if self.scheduler.state == STATE_RUNNING:
            scheduler_action_button = 'ENABLED'
            url_scheduler_action = url_for('tasks.xhr', node=self.node, action='disable')
        else:
            scheduler_action_button = 'DISABLED'
            url_scheduler_action = url_for('tasks.xhr', node=self.node, action='enable')

        self.kwargs = dict(
            node=self.node,
            tasks=tasks,
            url_add_task=url_for('schedule', node=self.node, add_task='True'),
            scheduler_action_button=scheduler_action_button,
            url_scheduler_action=url_scheduler_action,
            url_tasks_history=url_for('tasks.history')
        )

    def remove_apscheduler_job_without_task(self):
        # In case the task is remove from database while its apscheduler_job is still running
        apscheduler_job_id_set = set([j.id for j in self.scheduler.get_jobs(jobstore='default')])  # type(j.id): str
        task_id_set = set([str(t.id) for t in Task.query.all()])  # type(t.id): int
        for i in apscheduler_job_id_set.difference(task_id_set):
            self.scheduler.remove_job(i, jobstore='default')
            msg = "apscheduler_job #{id} removed since task #{id} not exist. ".format(id=i)
            apscheduler_logger.error(msg)
            flash(msg, self.WARN)

    def process_tasks(self, tasks):
        with db.session.no_autoflush:  # To avoid in place updating
            # for task in tasks:  # TypeError: 'Pagination' object is not iterable  # tasks.item: list
            for index, task in enumerate(tasks.items, (tasks.page - 1) * tasks.per_page + 1):
                # Columns: Name | Prev run result | Task results
                task.index = index
                task.name = task.name or ''
                task.timezone = task.timezone or self.scheduler.timezone
                task.create_time = self.remove_microsecond(task.create_time)
                task.update_time = self.remove_microsecond(task.update_time)
                task_results = TaskResult.query.filter_by(task_id=task.id).order_by(TaskResult.id.desc())
                task.run_times = task_results.count()
                task.url_task_results = url_for('tasks', node=self.node, task_id=task.id)
                if task.run_times > 0:
                    task.fail_times = sum([int(t.fail_count > 0) for t in task_results])
                    latest_task_result = task_results[0]
                    if latest_task_result.fail_count == 0 and latest_task_result.pass_count == 1:
                        task_job_result = TaskJobResult.query.filter_by(task_result_id=latest_task_result.id).order_by(
                            TaskJobResult.id.desc()).first()
                        task.prev_run_result = task_job_result.result[-19:]  # task_N_2019-01-01T00_00_01
                        task.url_prev_run_result = url_for('log', node=task_job_result.node, opt='stats',
                                                           project=task.project, spider=task.spider,
                                                           job=task_job_result.result)
                    else:
                        # 'FAIL 0, PASS 0' if execute_task() has not finished
                        task.prev_run_result = 'FAIL %s, PASS %s' % (latest_task_result.fail_count,
                                                                     latest_task_result.pass_count)
                        task.url_prev_run_result = url_for('tasks', node=self.node,
                                                           task_id=task.id, task_result_id=latest_task_result.id)
                else:
                    task.fail_times = 0
                    task.prev_run_result = self.NA
                    task.url_prev_run_result = task.url_task_results
                # Columns: Status | Actions | Next run time
                task.url_edit = url_for('schedule', node=self.node, task_id=task.id)
                # PostgreSQL 8.3 removes implicit casts
                # sqlalchemy.exc.ProgrammingError: (psycopg2.ProgrammingError) operator does not exist: character varying = integer
                # LINE 3: WHERE apscheduler_jobs.id = 2
                # HINT:  No operator matches the given name and argument types. You might need to add explicit type casts.
                # [SQL: 'SELECT apscheduler_jobs.job_state \nFROM apscheduler_jobs \nWHERE apscheduler_jobs.id = %(id_1)s'] [parameters: {'id_1': 2}]
                # (Background on this error at: http://sqlalche.me/e/f405)
                apscheduler_job = self.scheduler.get_job(str(task.id))  # Return type: Job or None
                if apscheduler_job:
                    self.logger.debug("apscheduler_job %s: %s", apscheduler_job.name, apscheduler_job)
                    if apscheduler_job.next_run_time:
                        task.status = 'Running'
                        action = 'pause'
                        if self.scheduler.state == STATE_PAUSED:
                            task.next_run_time = "Click DISABLED button first. "
                        else:
                            # TypeError: argument of type 'datetime.datetime' is not iterable
                            task.next_run_time = str(apscheduler_job.next_run_time)  # '2019-01-01 00:00:01+08:00'
                        task.url_fire = url_for('tasks.xhr', node=self.node, action='fire', task_id=task.id)
                    else:
                        task.status = 'Paused'
                        action = 'resume'
                        task.next_run_time = self.NA
                        task.url_fire = ''
                    task.url_status = url_for('tasks.xhr', node=self.node, action=action, task_id=task.id)
                    task.action = 'Stop'
                    task.url_action = url_for('tasks.xhr', node=self.node, action='remove', task_id=task.id)
                else:
                    task.status = 'Finished'
                    task.url_status = task.url_task_results  # '',  'javascript:;'
                    task.action = 'Delete'
                    task.url_action = url_for('tasks.xhr', node=self.node, action='delete', task_id=task.id)
                    task.next_run_time = self.NA
                    task.url_fire = ''

    def query_task_results(self):
        task_results = TaskResult.query.filter_by(task_id=self.task_id).order_by(
            TaskResult.id.desc()).paginate(page=self.page, per_page=self.per_page, error_out=False)
        # In case that execute_task() has not finished or selected_nodes is modified
        with_job = all([task_result.fail_count + task_result.pass_count == 1 for task_result in task_results.items])

        with db.session.no_autoflush:
            for index, task_result in enumerate(task_results.items,
                                                (task_results.page - 1) * task_results.per_page + 1):
                task_result.index = index
                if with_job:  # To show task_job_result in task_results.html
                    self.template = 'scrapydweb/task_results_with_job.html'
                    task_job_result = TaskJobResult.query.filter_by(task_result_id=task_result.id).order_by(
                        TaskJobResult.id.desc()).first()
                    task_result.task_job_result_id = task_job_result.id
                    task_result.run_time = self.remove_microsecond(task_job_result.run_time)
                    task_result.node = task_job_result.node
                    task_result.server = task_job_result.server
                    task_result.status_code = task_job_result.status_code
                    task_result.status = task_job_result.status
                    task_result.result = task_job_result.result
                    if task_job_result.status == self.OK:
                        task_result.url_stats = url_for('log', node=task_job_result.node, opt='stats',
                                                        project=self.task.project, spider=self.task.spider,
                                                        job=task_job_result.result)
                    else:
                        task_result.url_stats = ''  # 'javascript:;'
                else:
                    task_result.execute_time = self.remove_microsecond(task_result.execute_time)
                    task_result.url_task_job_results = url_for('tasks', node=self.node,
                                                               task_id=self.task_id, task_result_id=task_result.id)
                task_result.url_action = url_for('tasks.xhr', node=self.node, action='delete',
                                                 task_id=self.task.id, task_result_id=task_result.id)

        self.kwargs = dict(
            node=self.node,
            task_id=self.task_id,
            task=self.task,
            task_results=task_results,
            url_tasks=url_for('tasks', node=self.node),
        )

    def query_task_job_results(self):
        # https://docs.sqlalchemy.org/en/latest/core/sqlelement.html#sqlalchemy.sql.expression.asc
        task_job_results = TaskJobResult.query.filter_by(task_result_id=self.task_result_id).order_by(
            TaskJobResult.node.asc()).paginate(page=self.page, per_page=self.per_page, error_out=False)
        with db.session.no_autoflush:
            for index, task_job_result in enumerate(task_job_results.items,
                                                    (task_job_results.page - 1) * task_job_results.per_page + 1):
                task_job_result.index = index
                task_job_result.run_time = self.remove_microsecond(task_job_result.run_time)
                if task_job_result.status == self.OK:
                    task_job_result.url_stats = url_for('log', node=task_job_result.node, opt='stats',
                                                        project=self.task.project, spider=self.task.spider,
                                                        job=task_job_result.result)
                    task_job_result.url_clusterreports = url_for('clusterreports', node=self.node,
                                                                 project=self.task.project, spider=self.task.spider,
                                                                 job=task_job_result.result)
                else:
                    task_job_result.url_stats = ''  # 'javascript:;'
                    task_job_result.url_clusterreports = ''

        self.kwargs = dict(
            node=self.node,
            task_id=self.task_id,
            task_result_id=self.task_result_id,
            task=self.task,
            task_job_results=task_job_results,
            url_tasks=url_for('tasks', node=self.node),
            url_task_results=url_for('tasks', node=self.node, task_id=self.task_id),
        )


class TasksXhrView(BaseView):

    def __init__(self):
        super(TasksXhrView, self).__init__()

        self.action = self.view_args['action']  # pause|resume|remove|delete|dump|fire
        self.task_id = self.view_args['task_id']  # <int:task_id>
        self.task_result_id = self.view_args['task_result_id']  # <int:task_result_id>

        self.task = Task.query.get(self.task_id) if self.task_id else None
        self.apscheduler_job = self.scheduler.get_job(str(self.task_id)) if self.task_id else None  # Return type: Job|None
        self.js = dict(action=self.action, task_id=self.task_id, task_result_id=self.task_result_id, url=request.url)

    def dispatch_request(self, **kwargs):
        try:
            self.generate_response()
        except Exception as err:
            self.logger.error(traceback.format_exc())
            db.session.rollback()
            self.js['status'] = 'exception'
            self.js['message'] = str(err)
        else:
            self.js.setdefault('status', self.OK)
        finally:
            self.logger.debug(self.js)
            return self.json_dumps(self.js, as_response=True)

    def generate_response(self):
        if self.action in ['disable', 'enable']:  # ENABLE|DISABLE the scheduler
            self.enable_disable_scheduler()
        elif self.action == 'delete':  # delete a task_result|task
            if self.task_result_id:
                self.delete_task_result()
            else:
                self.delete_task()
        elif self.action == 'dump':  # For test only
            self.dump_task_data()
        elif self.action == 'fire':  # update next_run_time
            self.fire_task()
        elif self.action == 'list':  # For test only
            self.list_tasks_or_results()
        else:  # pause|resume|remove a apscheduler_job
            self.handle_apscheduler_job()

    def enable_disable_scheduler(self):
        # scheduler.running: a shortcut for scheduler.state != STATE_STOPPED
        # if self.scheduler.state == STATE_RUNNING:
        if self.action == 'disable':
            self.scheduler.pause()
        else:  # 'enable'
            self.scheduler.resume()
        handle_metadata('scheduler_state', self.scheduler.state)
        self.js['tip'] = "Scheduler after '%s': %s" % (self.action, SCHEDULER_STATE_DICT[self.scheduler.state])

    def delete_task_result(self):
        task_result = TaskResult.query.get(self.task_result_id)
        # In case that execute_task() has not finished
        # if task_result and (task_result.pass_count or task_result.fail_count):
        if task_result:
            db.session.delete(task_result)
            db.session.commit()
            self.js['tip'] = "task_result #%s deleted. " % self.task_result_id
        else:
            self.js['status'] = self.ERROR
            self.js['message'] = "task_result #%s not found. " % self.task_result_id

    def delete_task(self):
        # Actually, the 'delete a task' button is available only when  apscheduler_job is None
        if request.args.get('ignore_apscheduler_job') == 'True':  # For test only
            self.js['tip'] = "Ignore apscheduler_job #%s. " % self.task_id
        else:
            if self.apscheduler_job:
                self.apscheduler_job.remove()
                self.js['tip'] = "apscheduler_job #%s removed. " % self.task_id
            else:
                self.js['tip'] = "apscheduler_job #%s not found. " % self.task_id
        if self.task:
            db.session.delete(self.task)
            db.session.commit()
            msg = "Task #%s deleted. " % self.task_id
            apscheduler_logger.warning(msg)
            self.js['tip'] += msg
        else:
            self.js['status'] = self.ERROR
            self.js['message'] = self.js.pop('tip') + "Task #%s not found. " % self.task_id

    def fire_task(self):
        if not self.apscheduler_job:
            self.js['status'] = self.ERROR
            self.js['message'] = "apscheduler_job #{0} not found, check if task #{0} is finished. ".format(self.task_id)
            return
        elif not self.apscheduler_job.next_run_time:
            self.js['status'] = self.ERROR
            self.js['message'] = "apscheduler_job #%s is paused, resume it first. " % self.task_id
            return
        self.apscheduler_job.modify(next_run_time=datetime.now())
        self.js['tip'] = "Reload this page several seconds later to check out the fire result. "
        self.js['url_jump'] = url_for('tasks', node=self.node, task_id=self.task_id)

    def handle_apscheduler_job(self):
        if not self.apscheduler_job:
            self.js['status'] = self.ERROR
            self.js['message'] = "apscheduler_job #%s not found. " % self.task_id
            return
        if self.action == 'pause':
            self.apscheduler_job.pause()
        elif self.action == 'resume':
            self.apscheduler_job.resume()
        else:  # 'Stop' button to 'remove'
            self.apscheduler_job.remove()
        self.js['tip'] = u"apscheduler_job #{task_id} after '{action}': {apscheduler_job}".format(
            task_id=self.task_id, action=self.action, apscheduler_job=self.scheduler.get_job(str(self.task_id)))

    def dump_task_data(self):
        if not self.task:
            self.js['status'] = self.ERROR
            if self.apscheduler_job:  # For test only
                self.js['data'] = dict(apscheduler_job=self.task_id)
                self.js['message'] = "apscheduler_job #%s found. " % self.task_id
            else:
                self.js['data'] = None
                self.js['message'] = "apscheduler_job #%s not found. " % self.task_id
            self.js['message'] += "Task #%s not found. " % self.task_id
            return
        # print(vars(self.task))
        self.js['data'] = dict((k, v) for k, v in vars(self.task).items() if not k.startswith('_'))
        self.js['data']['settings_arguments'] = json.loads(self.js['data']['settings_arguments'])
        self.js['data']['selected_nodes'] = json.loads(self.js['data']['selected_nodes'])
        self.js['data']['create_time'] = str(self.js['data']['create_time'])
        self.js['data']['update_time'] = str(self.js['data']['update_time'])
        if not self.apscheduler_job:
            self.js['data']['apscheduler_job'] = None
            self.js['tip'] = "apscheduler_job #{id} not found. Task #{id} found. ".format(id=self.task_id)
            return
        # print(self.apscheduler_job.__slots__)
        # ('_scheduler', '_jobstore_alias', 'id', 'trigger', 'executor', 'func', 'func_ref', 'args', 'kwargs',
        #  'name', 'misfire_grace_time', 'coalesce', 'max_instances', 'next_run_time')
        self.js['data']['apscheduler_job'] = dict(
            id=self.apscheduler_job.id,
            name=self.apscheduler_job.name,
            kwargs=self.apscheduler_job.kwargs,
            misfire_grace_time=self.apscheduler_job.misfire_grace_time,
            coalesce=self.apscheduler_job.coalesce,
            max_instances=self.apscheduler_job.max_instances,
            next_run_time=str(self.apscheduler_job.next_run_time) if self.apscheduler_job.next_run_time else None,
        )
        self.js['data']['apscheduler_job']['trigger'] = dict((f.name, str(f))
                                                             for f in self.apscheduler_job.trigger.fields)
        start_date = self.apscheduler_job.trigger.start_date
        self.js['data']['apscheduler_job']['trigger'].update(dict(
            start_date=str(start_date) if start_date else None,
            end_date=str(self.apscheduler_job.trigger.end_date) if self.apscheduler_job.trigger.end_date else None,
            timezone=str(self.apscheduler_job.trigger.timezone) if self.apscheduler_job.trigger.timezone else None,
            jitter=self.apscheduler_job.trigger.jitter,
        ))
        self.js['tip'] = "apscheduler_job #{id} found. Task #{id} found. ".format(id=self.task_id)

    def list_tasks_or_results(self):
        if self.task_id and self.task_result_id:
            records = TaskJobResult.query.filter_by(task_result_id=self.task_result_id).all()
        elif self.task_id:
            records = TaskResult.query.filter_by(task_id=self.task_id).all()
        else:
            records = Task.query.all()
        self.js['ids'] = [i.id for i in records]



# coding: utf-8
import platform
import re

from scrapy import __version__ as scrapy_version

from tests.utils import cst, req, sleep, switch_scrapyd, upload_file_deploy


NODE = 2
FILENAME = '%s_%s_%s.pickle' % (cst.PROJECT, cst.VERSION, cst.SPIDER)
KEY = '/'.join([cst.PROJECT, cst.SPIDER, cst.JOBID])
run_data = {
    '1': 'on',
    '2': 'on',
    'checked_amount': '2',
    'filename': FILENAME
}
run_data_single_scrapyd = {
    '1': 'on',
    'checked_amount': '1',
    'filename': FILENAME
}


# Multinode Run Spider button in deploy results page
# Multinode Run Spider button in servers page
def test_schedule_from_post(app, client):
    text, __ = req(app, client, view='schedule', kws=dict(node=NODE), data={'1': 'on', '2': 'on'})
    assert (re.search(r'id="checkbox_1".*?checked.*?/>', text, re.S)
            and re.search(r'id="checkbox_2".*?checked.*?/>', text, re.S))


# CHECK first to generate xx.pickle for RUN
def test_check(app, client):
    upload_file_deploy(app, client, filename='ScrapydWeb_demo.egg', project=cst.PROJECT, redirect_project=cst.PROJECT)
    data = dict(
        project=cst.PROJECT,
        _version=cst.VERSION,
        spider=cst.SPIDER,
        jobid=cst.JOBID,
        additional="-d setting=LOGSTATS_INTERVAL=10"  # For the test_telnet_in_stats() below

    )
    req(app, client, view='schedule.check', kws=dict(node=NODE), data=data,
        jskws=dict(filename=FILENAME))


def test_run(app, client):
    node = 1

    req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data,
        ins=['run results - ScrapydWeb', 'id="checkbox_1"', 'id="checkbox_2"', 'onclick="passToServers();"'])

    # test handle_unique_constraint() in jobs.py
    sleep()
    req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data_single_scrapyd,
        ins=['run results - ScrapydWeb', 'id="checkbox_1"', 'onclick="passToServers();"'], nos='id="checkbox_2"')
    keep_text = ''
    for times in [1, 2]:
        __, js = req(app, client, view='api', kws=dict(node=node, opt='listjobs', project=cst.PROJECT))
        if js['pending']:
            final_pending_job = js['pending'][-1]
            assert final_pending_job['id'] == cst.JOBID
            first_job = js['running'][-1]
            first_job_start = first_job['start_time'][:19]
            # Ignore seen pending job: ScrapydWeb_demo/test/2018-01-01T01_01_02
            flash = "Ignore seen pending job: %s" % KEY
            ins = ["Vue.extend(Main)", "start: '%s'," % first_job_start]
            nos = ['class="table wrap"', "Ignore seen running job"]
            text, __ = req(app, client, view='jobs', kws=dict(node=node, style='database'))
            try:
                if times == 1:
                    assert flash in text
                else:
                    assert flash not in text
                for i in ins:
                    assert i in text
                for n in nos:
                    assert n not in text
            except AssertionError:
                # the response containS 'Ignore seen running' instead
                keep_text = text
                break
        else:
            break

    sleep()
    __, js = req(app, client, view='api', kws=dict(node=node, opt='listjobs', project=cst.PROJECT))
    first_job = js['running'][-2]
    first_job_start = first_job['start_time'][:19]
    second_job = js['running'][-1]
    second_job_start = second_job['start_time'][:19]
    assert first_job['id'] == second_job['id'] == cst.JOBID
    # TODO: For compatibility with Debian?! Running job with same key is not ordered by start ASC?!
    # assert second_job_start > first_job_start
    # Ignore seen running job: ScrapydWeb_demo/test/2018-01-01T01_01_02, started at 2019-03-01 20:27:22
    flash = "Ignore seen running job: %s, started at %s" % (KEY, first_job_start)
    if keep_text:
        text = keep_text
    else:
        text, __ = req(app, client, view='jobs', kws=dict(node=node, style='database'))
    for i in [flash, "Vue.extend(Main)", "start: '%s'," % second_job_start]:
        assert i in text
    for n in ['class="table wrap"', "start: '%s'," % first_job_start]:
        assert n not in text
    # flash only once
    req(app, client, view='jobs', kws=dict(node=node, style='database'),
        ins=["Vue.extend(Main)", "start: '%s'," % second_job_start],
        nos=[flash, 'class="table wrap"', "start: '%s'," % first_job_start])

    for i in range(2):
        req(app, client, view='api',
            kws=dict(node=node, opt='forcestop', project=cst.PROJECT, version_spider_job=cst.JOBID))
    sleep()
    __, js = req(app, client, view='api', kws=dict(node=node, opt='listjobs', project=cst.PROJECT))
    last_but_two_finished_job = js['finished'][-2]
    last_but_two_finished_job_start = last_but_two_finished_job['start_time'][:19]
    last_finished_job = js['finished'][-1]
    last_finished_job_start = last_finished_job['start_time'][:19]
    assert last_but_two_finished_job['id'] == last_finished_job['id'] == cst.JOBID
    # Ignore seen finished job: ScrapydWeb_demo/test/2018-01-01T01_01_02, started at 2019-03-01 20:27:22
    flash = "Ignore seen finished job: %s, started at %s" % (KEY, last_but_two_finished_job_start)
    req(app, client, view='jobs', kws=dict(node=node, style='database'),
        ins=[flash, "Vue.extend(Main)", "start: '%s'," % last_finished_job_start],
        nos=['class="table wrap"', "start: '%s'," % last_but_two_finished_job_start])
    # flash only once
    req(app, client, view='jobs', kws=dict(node=node, style='database'),
        ins=["Vue.extend(Main)", "start: '%s'," % last_finished_job_start],
        nos=[flash, 'class="table wrap"', "start: '%s'," % last_but_two_finished_job_start])


# Note that in LogParser is enabled in test_enable_logparser(), with PARSE_ROUND_INTERVAL defaults to 10.
# And LOGSTATS_INTERVAL is set to 10 in test_check() above.
# This test would fail if Scrapy >= 1.5.2 since telnet console now requires username and password
# https://doc.scrapy.org/en/latest/news.html#scrapy-1-5-2-2019-01-22
def test_telnet_in_stats(app, client):
    node = 1
    desktop_ins = [">Log analysis</li>", ">Log categorization</li>", ">View log</li>", ">Progress visualization</li>"]
    mobile_ins = [">Analysis</li>", ">Categories</li>", ">Charts</li>", ">Logs</li>"]
    telnet_ins = [">Crawler.stats</li>", "<td>datetime.datetime(",
                  ">Crawler.engine</li>", "<th>engine.scraper.is_idle()</th>", "<td>telnet</td>"]
    telnet_nos = ["CRITICAL: Unhandled Error", "telnet.OptionRefused"]
    req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data, ins="run results - ScrapydWeb")

    kws = dict(node=node, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.JOBID)
    for i in range(1, 4):
        sleep(10)
        print(i * 10)
        text, __ = req(app, client, view='log', kws=kws)
        if desktop_ins[-1] in text and telnet_ins[-1] in text:
            print("Found: %s %s" % (desktop_ins[-1], telnet_ins[-1]))
            break
    # test jobs POST data={} to save pages and items in database
    __, js = req(app, client, view='jobs', kws=dict(node=node), data={})
    assert isinstance(js[KEY]['pages'], int)  # and js[KEY]['pages'] > 0

    # Linux-5.0.9-301.fc30.x86_64-x86_64-with-fedora-30-Thirty'
    if (platform.system() == 'Windows' or 'fedora' in platform.platform()) and scrapy_version > '1.5.1':
        print("telnet not available for scrapy_version: %s" % scrapy_version)
        telnet_ins = []

    req(app, client, view='log', kws=kws, ins=desktop_ins + telnet_ins, nos=telnet_nos)

    kws.update(ui='mobile')
    req(app, client, view='log', kws=kws,
        ins=mobile_ins + telnet_ins, nos=telnet_nos, mobileui=True)

    req(app, client, view='api',
        kws=dict(node=node, opt='forcestop', project=cst.PROJECT, version_spider_job=cst.JOBID))


def test_pending_jobs(app, client):
    node = 1
    for i in range(2):
        req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data, ins="run results - ScrapydWeb")
    req(app, client, view='jobs', kws=dict(node=node, style='database'),
        ins="Vue.extend(Main)", nos='class="table wrap"')
    sleep()
    for i in range(2):
        req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data, ins="run results - ScrapydWeb")
    req(app, client, view='jobs', kws=dict(node=node, style='database'),
        ins="Vue.extend(Main)", nos='class="table wrap"')
    sleep()
    req(app, client, view='jobs', kws=dict(node=node, style='database'),
        ins=["Ignore seen running job: %s, started at" % KEY, "Vue.extend(Main)"], nos='class="table wrap"')
    for i in range(4):
        req(app, client, view='api',
            kws=dict(node=node, opt='forcestop', project=cst.PROJECT, version_spider_job=cst.JOBID))


def test_run_fail(app, client):
    switch_scrapyd(app)
    req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data, ins='Multinode schedule terminated')


def test_schedule_xhr(app, client):
    req(app, client, view='schedule.xhr',
        kws=dict(node=NODE, filename=FILENAME),
        jskws=dict(status=cst.ERROR))



# coding: utf-8
"""
How ScrapydWeb works:
BROWSER <<<>>> SCRAPYDWEB_BIND:SCRAPYDWEB_PORT <<<>>> your SCRAPYD_SERVERS

GitHub: https://github.com/my8100/scrapydweb
DOCS: https://github.com/my8100/files/blob/master/scrapydweb/README.md
文档：https://github.com/my8100/files/blob/master/scrapydweb/README_CN.md
"""
import os


############################## QUICK SETUP start ##############################
############################## 快速设置 开始 ###################################
# Setting SCRAPYDWEB_BIND to '0.0.0.0' or IP-OF-THE-CURRENT-HOST would make
# ScrapydWeb server visible externally; Otherwise, set it to '127.0.0.1'.
# The default is '0.0.0.0'.
SCRAPYDWEB_BIND = '0.0.0.0'
# Accept connections on the specified port, the default is 5000.
SCRAPYDWEB_PORT = 5000

# The default is False, set it to True to enable basic auth for the web UI.
ENABLE_AUTH = False
# In order to enable basic auth, both USERNAME and PASSWORD should be non-empty strings.
USERNAME = ''
PASSWORD = ''


# Make sure that [Scrapyd](https://github.com/scrapy/scrapyd) has been installed
# and started on all of your hosts.
# Note that for remote access, you have to manually set 'bind_address = 0.0.0.0'
# in the configuration file of Scrapyd and restart Scrapyd to make it visible externally.
# Check out 'https://scrapyd.readthedocs.io/en/latest/config.html#example-configuration-file' for more info.
# ------------------------------ Chinese --------------------------------------
# 请先确保所有主机都已经安装和启动 [Scrapyd](https://github.com/scrapy/scrapyd)。
# 如需远程访问 Scrapyd，则需在 Scrapyd 配置文件中设置 'bind_address = 0.0.0.0'，然后重启 Scrapyd。
# 详见 https://scrapyd.readthedocs.io/en/latest/config.html#example-configuration-file

# - the string format: username:password@ip:port#group
#   - The default port would be 6800 if not provided,
#   - Both basic auth and group are optional.
#   - e.g. '127.0.0.1:6800' or 'username:password@localhost:6801#group'
# - the tuple format: (username, password, ip, port, group)
#   - When the username, password, or group is too complicated (e.g. contains ':@#'),
#   - or if ScrapydWeb fails to parse the string format passed in,
#   - it's recommended to pass in a tuple of 5 elements.
#   - e.g. ('', '', '127.0.0.1', '6800', '') or ('username', 'password', 'localhost', '6801', 'group')
SCRAPYD_SERVERS = [
    '127.0.0.1:6800',
    # 'username:password@localhost:6801#group',
    ('username', 'password', 'localhost', '6801', 'group'),
]


# It's recommended to update the three options below
# if both ScrapydWeb and one of your Scrapyd servers run on the same machine.
# ------------------------------ Chinese --------------------------------------
# 假如 ScrapydWeb 和某个 Scrapyd 运行于同一台主机，建议更新如下三个设置项。

# If both ScrapydWeb and one of your Scrapyd servers run on the same machine,
# ScrapydWeb would try to directly read Scrapy logfiles from disk, instead of making a request
# to the Scrapyd server.
# e.g. '127.0.0.1:6800' or 'localhost:6801', do not forget the port number.
LOCAL_SCRAPYD_SERVER = ''

# Enter the directory when you run Scrapyd, run the command below
# to find out where the Scrapy logs are stored:
# python -c "from os.path import abspath, isdir; from scrapyd.config import Config; path = abspath(Config().get('logs_dir')); print(path); print(isdir(path))"
# Check out https://scrapyd.readthedocs.io/en/stable/config.html#logs-dir for more info.
# e.g. 'C:/Users/username/logs' or '/home/username/logs'
LOCAL_SCRAPYD_LOGS_DIR = ''

# The default is False, set it to True to automatically run LogParser as a subprocess at startup.
# Note that you can run the LogParser service separately via command 'logparser' as you like.
# Run 'logparser -h' to find out the config file of LogParser for more advanced settings.
# Visit https://github.com/my8100/logparser for more info.
ENABLE_LOGPARSER = False
############################## QUICK SETUP end ################################
############################## 快速设置 结束 ###################################


############################## ScrapydWeb #####################################
# The default is False, set it to True and add both CERTIFICATE_FILEPATH and PRIVATEKEY_FILEPATH
# to run ScrapydWeb in HTTPS mode.
# Note that this feature is not fully tested, please leave your comment here if ScrapydWeb
# raises any excepion at startup: https://github.com/my8100/scrapydweb/issues/18
ENABLE_HTTPS = False
# e.g. '/home/username/cert.pem'
CERTIFICATE_FILEPATH = ''
# e.g. '/home/username/cert.key'
PRIVATEKEY_FILEPATH = ''


############################## Scrapy #########################################
# ScrapydWeb is able to locate projects in the SCRAPY_PROJECTS_DIR,
# so that you can simply select a project to deploy, instead of packaging it in advance.
# e.g. 'C:/Users/username/myprojects' or '/home/username/myprojects'
SCRAPY_PROJECTS_DIR = ''


############################## Scrapyd ########################################
# ScrapydWeb would try every extension in sequence to locate the Scrapy logfile.
# The default is ['.log', '.log.gz', '.txt'].
SCRAPYD_LOG_EXTENSIONS = ['.log', '.log.gz', '.txt']

# The default is None, only set it up when you need to visit Scrapyd servers via reverse proxy.
# Make sure that SCRAPYD_SERVERS_PUBLIC_URLS has same length with SCRAPYD_SERVERS above.
# e.g.
# SCRAPYD_SERVERS_PUBLIC_URLS = [
    # 'https://a.b.com',  # visit the first Scrapyd server via reverse proxy.
    # '',  # visit the second Scrapyd server without reverse proxy.
# ]
# See https://github.com/my8100/scrapydweb/issues/94 for more info.
SCRAPYD_SERVERS_PUBLIC_URLS = None


############################## LogParser ######################################
# Whether to backup the stats json files locally after you visit the Stats page of a job
# so that it is still accessible even if the original logfile has been deleted.
# The default is True, set it to False to disable this behaviour.
BACKUP_STATS_JSON_FILE = True


############################## Timer Tasks ####################################
# Run ScrapydWeb with argument '-sw' or '--switch_scheduler_state', or click the ENABLED|DISABLED button
# on the Timer Tasks page to turn on/off the scheduler for the timer tasks and the snapshot mechanism below.

# The default is 300, which means ScrapydWeb would automatically create a snapshot of the Jobs page
# and save the jobs info in the database in the background every 300 seconds.
# Note that this behaviour would be paused if the scheduler for timer tasks is disabled.
# Set it to 0 to disable this behaviour.
JOBS_SNAPSHOT_INTERVAL = 300


############################## Run Spider #####################################
# The default is False, set it to True to automatically
# expand the 'settings & arguments' section in the Run Spider page.
SCHEDULE_EXPAND_SETTINGS_ARGUMENTS = False

# The default is 'Mozilla/5.0', set it a non-empty string to customize the default value of `custom`
# in the drop-down list of `USER_AGENT`.
SCHEDULE_CUSTOM_USER_AGENT = 'Mozilla/5.0'

# The default is None, set it to any value of ['custom', 'Chrome', 'iPhone', 'iPad', 'Android']
# to customize the default value of `USER_AGENT`.
SCHEDULE_USER_AGENT = None

# The default is None, set it to True or False to customize the default value of `ROBOTSTXT_OBEY`.
SCHEDULE_ROBOTSTXT_OBEY = None

# The default is None, set it to True or False to customize the default value of `COOKIES_ENABLED`.
SCHEDULE_COOKIES_ENABLED = None

# The default is None, set it to a non-negative integer to customize the default value of `CONCURRENT_REQUESTS`.
SCHEDULE_CONCURRENT_REQUESTS = None

# The default is None, set it to a non-negative number to customize the default value of `DOWNLOAD_DELAY`.
SCHEDULE_DOWNLOAD_DELAY = None

# The default is "-d setting=CLOSESPIDER_TIMEOUT=60\r\n-d setting=CLOSESPIDER_PAGECOUNT=10\r\n-d arg1=val1",
# set it to '' or any non-empty string to customize the default value of `additional`.
# Use '\r\n' as the line separator.
SCHEDULE_ADDITIONAL = "-d setting=CLOSESPIDER_TIMEOUT=60\r\n-d setting=CLOSESPIDER_PAGECOUNT=10\r\n-d arg1=val1"


############################## Page Display ###################################
# The default is True, set it to False to hide the Items page, as well as
# the Items column in the Jobs page.
SHOW_SCRAPYD_ITEMS = True

# The default is True, set it to False to hide the Job column in the Jobs page with non-database view.
SHOW_JOBS_JOB_COLUMN = True

# The default is 0, which means unlimited, set it to a positive integer so that
# only the latest N finished jobs would be shown in the Jobs page with non-database view.
JOBS_FINISHED_JOBS_LIMIT = 0

# If your browser stays on the Jobs page, it would be reloaded automatically every N seconds.
# The default is 300, set it to 0 to disable auto-reloading.
JOBS_RELOAD_INTERVAL = 300

# The load status of the current Scrapyd server is checked every N seconds,
# which is displayed in the top right corner of the page.
# The default is 10, set it to 0 to disable auto-refreshing.
DAEMONSTATUS_REFRESH_INTERVAL = 10


############################## Send Text ######################################
########## usage in scrapy projects ##########
# See the "Send Text" page

########## slack ##########
# How to create a slack app:
# 1. Visit https://api.slack.com/apps and press the "Create New App" button.
# 2. Enter your App Name (e.g. myapp)and select one of your Slack Workspaces, the press "Create App".
# 3. Click the "OAuth & Permissions" menu in the sidebar on the left side of the page.
# 4. Scroll down the page and find out "Select Permission Scopes" in the "Scopes" section
# 5. Enter "send" and select "Send messages as <your-app-name>", then press "Save Changes"
# 6. Scroll up the page and press "Install App to Workspace", then press "Install"
# 7. Copy the "OAuth Access Token", e.g. xoxp-123-456-789-abcde
# See https://api.slack.com/apps for more info

# See step 1~7 above, e.g. 'xoxp-123-456-789-abcde'
SLACK_TOKEN = os.environ.get('SLACK_TOKEN', '')
# The default channel to use when sending text via slack, e.g. 'general'
SLACK_CHANNEL = 'general'

########## telegram ##########
# How to create a telegram bot:
# 1. Visit https://telegram.me/botfather to start a conversation with Telegram's bot that creates other bots.
# 2. Send the /newbot command to create a new bot in a chat with BotFather.
# 3. Follow the instructions to set up name and username (e.g. my_bot) for your bot.
# 4. You would get a token (e.g. 123:abcde) after step 3.
# 5. Visit telegram.me/<bot_username> (e.g. telegram.me/my_bot) and say hi to your bot to initiate a conversation.
# 6. Visit https://api.telegram.org/bot<token-in-setp-4>/getUpdates to get the chat_id.
#    (e.g. Visit https://api.telegram.org/bot123:abcde/getUpdates
#     and you can find the chat_id in "chat":{"id":123456789,...)
# See https://core.telegram.org/bots#6-botfather for more info

# See step 1~4 above, e.g. '123:abcde'
TELEGRAM_TOKEN = os.environ.get('TELEGRAM_TOKEN', '')
# See step 5~6 above, e.g. 123456789
TELEGRAM_CHAT_ID = int(os.environ.get('TELEGRAM_CHAT_ID', 0))

########## email ##########
# The default subject to use when sending text via email.
EMAIL_SUBJECT = 'Email from #scrapydweb'

########## email sender & recipients ##########
# Leave this option as '' to default to the EMAIL_SENDER option below; Otherwise, set it up
# if your email service provider requires an username which is different from the EMAIL_SENDER option below to login.
# e.g. 'username'
EMAIL_USERNAME = ''
# As for different email service provider, you might have to get an APP password (like Gmail)
# or an authorization code (like QQ mail) and set it as the EMAIL_PASSWORD.
# Check out links below to get more help:
# https://stackoverflow.com/a/27515833/10517783 How to send an email with Gmail as the provider using Python?
# https://stackoverflow.com/a/26053352/10517783 Python smtplib proxy support
# e.g. 'password4gmail'
EMAIL_PASSWORD = os.environ.get('EMAIL_PASSWORD', '')

# e.g. 'username@gmail.com'
EMAIL_SENDER = ''
# e.g. ['username@gmail.com', ]
EMAIL_RECIPIENTS = [EMAIL_SENDER]

########## email smtp settings ##########
# Check out this link if you are using ECS of Alibaba Cloud and your SMTP server provides TCP port 25 only:
# https://www.alibabacloud.com/help/doc-detail/56130.htm
# Config for https://mail.google.com using SSL: ('smtp.gmail.com', 465, True)
# Config for https://mail.google.com:           ('smtp.gmail.com', 587, False)
# Config for https://mail.qq.com using SSL:     ('smtp.qq.com', 465, True)
# Config for http://mail.10086.cn:              ('smtp.139.com', 25, False)
SMTP_SERVER = ''
SMTP_PORT = 0
SMTP_OVER_SSL = False
# The timeout in seconds for the connection attempt, the default is 30.
SMTP_CONNECTION_TIMEOUT = 30


############################## Monitor & Alert ################################
# The default is False, set it to True to launch the poll subprocess to monitor your crawling jobs.
ENABLE_MONITOR = False

########## poll interval ##########
# Tip: In order to be notified (and stop or forcestop a job when triggered) in time,
# you can reduce the value of POLL_ROUND_INTERVAL and POLL_REQUEST_INTERVAL,
# at the cost of burdening both CPU and bandwidth of your servers.

# Sleep N seconds before starting next round of poll, the default is 300.
POLL_ROUND_INTERVAL = 300
# Sleep N seconds between each request to the Scrapyd server while polling, the default is 10.
POLL_REQUEST_INTERVAL = 10

########## alert switcher ##########
# Tip: Set the SCRAPYDWEB_BIND option the in "QUICK SETUP" section to the actual IP of your host,
# then you can visit ScrapydWeb via the links attached in the alert.

# The default is False, set it to True to enable alert via Slack, Telegram, or Email.
# You have to set up your accounts in the "Send text" section above first.
ENABLE_SLACK_ALERT = False
ENABLE_TELEGRAM_ALERT = False
ENABLE_EMAIL_ALERT = False

########## alert working time ##########
# Monday is 1 and Sunday is 7.
# e.g, [1, 2, 3, 4, 5, 6, 7]
ALERT_WORKING_DAYS = []

# From 0 to 23.
# e.g. [9] + list(range(15, 18)) >>> [9, 15, 16, 17], or range(24) for 24 hours
ALERT_WORKING_HOURS = []

########## basic triggers ##########
# Trigger alert every N seconds for each running job.
# The default is 0, set it to a positive integer to enable this trigger.
ON_JOB_RUNNING_INTERVAL = 0

# Trigger alert when a job is finished.
# The default is False, set it to True to enable this trigger.
ON_JOB_FINISHED = False

########## advanced triggers ##########
# - LOG_XXX_THRESHOLD:
#   - Trigger alert the first time reaching the threshold for a specific kind of log.
#   - The default is 0, set it to a positive integer to enable this trigger.
# - LOG_XXX_TRIGGER_STOP (optional):
#   - The default is False, set it to True to stop current job automatically when reaching the LOG_XXX_THRESHOLD.
#   - The SIGTERM signal would be sent only one time to shut down the crawler gracefully.
#   - In order to avoid an UNCLEAN shutdown, the 'STOP' action would be executed one time at most
#   - if none of the 'FORCESTOP' triggers is enabled, no matter how many 'STOP' triggers are enabled.
# - LOG_XXX_TRIGGER_FORCESTOP (optional):
#   - The default is False, set it to True to FORCESTOP current job automatically when reaching the LOG_XXX_THRESHOLD.
#   - The SIGTERM signal would be sent twice resulting in an UNCLEAN shutdown, without the Scrapy stats dumped!
#   - The 'FORCESTOP' action would be executed if both of the 'STOP' and 'FORCESTOP' triggers are enabled.

# Note that the 'STOP' action and the 'FORCESTOP' action would still be executed even when the current time
# is NOT within the ALERT_WORKING_DAYS and the ALERT_WORKING_HOURS, though no alert would be sent.

LOG_CRITICAL_THRESHOLD = 0
LOG_CRITICAL_TRIGGER_STOP = False
LOG_CRITICAL_TRIGGER_FORCESTOP = False

LOG_ERROR_THRESHOLD = 0
LOG_ERROR_TRIGGER_STOP = False
LOG_ERROR_TRIGGER_FORCESTOP = False

LOG_WARNING_THRESHOLD = 0
LOG_WARNING_TRIGGER_STOP = False
LOG_WARNING_TRIGGER_FORCESTOP = False

LOG_REDIRECT_THRESHOLD = 0
LOG_REDIRECT_TRIGGER_STOP = False
LOG_REDIRECT_TRIGGER_FORCESTOP = False

LOG_RETRY_THRESHOLD = 0
LOG_RETRY_TRIGGER_STOP = False
LOG_RETRY_TRIGGER_FORCESTOP = False

LOG_IGNORE_THRESHOLD = 0
LOG_IGNORE_TRIGGER_STOP = False
LOG_IGNORE_TRIGGER_FORCESTOP = False


############################## System #########################################
# The default is False, set it to True to enable debug mode and the interactive debugger
# would be shown in the browser instead of the "500 Internal Server Error" page.
# Note that use_reloader is set to False in run.py
DEBUG = False

# The default is False, set it to True to change the logging level from INFO to DEBUG
# for getting more information about how ScrapydWeb works, especially while debugging.
VERBOSE = False

# The default is '', which means saving all program data in the Python directory.
# e.g. 'C:/Users/username/scrapydweb_data' or '/home/username/scrapydweb_data'
DATA_PATH = os.environ.get('DATA_PATH', '')

# The default is '', which means saving data of Jobs and Timer Tasks in DATA_PATH using SQLite.
# The data could be also saved in MySQL or PostgreSQL backend in order to improve concurrency.
# To use MySQL backend, run command: pip install --upgrade pymysql
# To use PostgreSQL backend, run command: pip install --upgrade psycopg2
# e.g.
# 'mysql://username:password@127.0.0.1:3306'
# 'postgresql://username:password@127.0.0.1:5432'
# 'sqlite:///C:/Users/username'
# 'sqlite:////home/username'
DATABASE_URL = os.environ.get('DATABASE_URL', '')



# coding: utf-8
from functools import partial
from io import BytesIO
import os
import re

from tests.utils import cst, req, switch_scrapyd, upload_file_deploy


def test_deploy_from_post(app, client):
    text, __ = req(app, client, view='deploy', kws=dict(node=1), data={'1': 'on', '2': 'on'})
    assert (re.search(r'id="checkbox_1".*?checked.*?/>', text, re.S)
            and re.search(r'id="checkbox_2".*?checked.*?/>', text, re.S))


def test_auto_packaging_select_option(app, client):
    ins = [
        '(14 projects)',
        u"var folders = ['demo - 副本', 'demo',",
        "var projects = ['demo-copy', 'demo',",
        '<div>%s<' % cst.PROJECT,
        u'<div>demo - 副本<',
        '<div>demo<',
        '<div>demo_only_scrapy_cfg<'
    ]
    nos = ['<div>demo_without_scrapy_cfg<']
    if not os.environ.get('DATA_PATH', ''):
        nos.append('<h3>No projects found')
    req(app, client, view='deploy', kws=dict(node=2), ins=ins, nos=nos)


# {'status': 'error', 'message': 'Traceback
# ...TypeError:...activate_egg(eggpath)...\'tuple\' object is not an iterator\r\n'}
def test_addversion(app, client):
    data = {
        '1': 'on',
        'checked_amount': '1',
        'project': 'fakeproject',
        'version': 'fakeversion',
        'file': (BytesIO(b'my file contents'), "fake.egg")
    }
    req(app, client, view='deploy.upload', kws=dict(node=2), data=data, ins='activate_egg')


# <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
# <title>Redirecting...</title>
# <h1>Redirecting...</h1>
# <p>You should be redirected automatically to target URL:
# <a href="/1/schedule/demo/2018-01-01T01_01_01/">/1/schedule/demo/2018-01-01T01_01_01/</a>.  If not click the link.
def test_auto_packaging(app, client):
    data = {
        '1': 'on',
        'checked_amount': '1',
        'folder': cst.PROJECT,
        'project': cst.PROJECT,
        'version': cst.VERSION
    }
    req(app, client, view='deploy.upload', kws=dict(node=2), data=data,
        ins=['deploy results - ScrapydWeb', 'onclick="multinodeRunSpider();"', 'id="checkbox_1"'],
        nos='id="checkbox_2"')

    data.update({'2': 'on', 'checked_amount': '2'})
    req(app, client, view='deploy.upload', kws=dict(node=2), data=data,
        ins=['deploy results - ScrapydWeb', 'onclick="multinodeRunSpider();"', 'id="checkbox_1"', 'id="checkbox_2"'])


def test_auto_packaging_unicode(app, client):
    if cst.WINDOWS_NOT_CP936:
        return
    data = {
        '1': 'on',
        'checked_amount': '1',
        'folder': u'demo - 副本',
        'project': u'demo - 副本',
        'version': cst.VERSION,
    }
    req(app, client, view='deploy.upload', kws=dict(node=2), data=data, ins=['deploy results', 'demo_____'])


def test_scrapy_cfg(app, client):
    for folder, result in cst.SCRAPY_CFG_DICT.items():
        data = {
            '1': 'on',
            '2': 'on',
            'checked_amount': '2',
            'folder': folder,
            'project': cst.PROJECT,
            'version': cst.VERSION,
        }
        ins = ['fail - ScrapydWeb', result] if result else 'deploy results - ScrapydWeb'
        req(app, client, view='deploy.upload', kws=dict(node=2), data=data, ins=ins)


def test_scrapy_cfg_first_node_not_exist(app, client):
    switch_scrapyd(app)
    for folder, result in cst.SCRAPY_CFG_DICT.items():
        data = {
            '1': 'on',
            '2': 'on',
            'checked_amount': '2',
            'folder': folder,
            'project': cst.PROJECT,
            'version': cst.VERSION,
        }
        nos = []
        if folder == 'demo_only_scrapy_cfg' or not result:
            ins = ['fail - ScrapydWeb', 'the first selected node returned status']
        else:
            ins = ['fail - ScrapydWeb', result]
            nos = 'the first selected node returned status'
        req(app, client, view='deploy.upload', kws=dict(node=2), data=data, ins=ins, nos=nos)


def test_upload_file_deploy(app, client):
    upload_file_deploy_multinode = partial(upload_file_deploy, app=app, client=client, multinode=True)

    filenames = ['demo.egg', 'demo_inner.zip', 'demo_outer.zip',
                 'demo - Win7CNsendzipped.zip', 'demo - Win10cp1252.zip']
    if cst.WINDOWS_NOT_CP936:
        filenames.extend(['demo - Ubuntu.zip', 'demo - Ubuntu.tar.gz', 'demo - macOS.zip', 'demo - macOS.tar.gz'])
    else:
        filenames.extend([u'副本.zip', u'副本.tar.gz', u'副本.egg', u'demo - 副本 - Win7CN.zip',
                          u'demo - 副本 - Win7CNsendzipped.zip', u'demo - 副本 - Win10cp936.zip',
                          u'demo - 副本 - Ubuntu.zip', u'demo - 副本 - Ubuntu.tar.gz',
                          u'demo - 副本 - macOS.zip', u'demo - 副本 - macOS.tar.gz'])

    for filename in filenames:
        if filename == 'demo.egg':
            project = cst.PROJECT
            redirect_project = cst.PROJECT
        else:
            project = re.sub(r'\.egg|\.zip|\.tar\.gz', '', filename)
            project = 'demo_unicode' if project == u'副本' else project
            redirect_project = re.sub(cst.STRICT_NAME_PATTERN, '_', project)
        upload_file_deploy_multinode(filename=filename, project=project, redirect_project=redirect_project)

    for filename, alert in cst.SCRAPY_CFG_DICT.items():
        if alert:
            upload_file_deploy_multinode(filename='%s.zip' % filename, project=filename, alert=alert, fail=True)
        else:
            upload_file_deploy_multinode(filename='%s.zip' % filename, project=filename, redirect_project=filename)

    switch_scrapyd(app)

    for filename, alert in cst.SCRAPY_CFG_DICT.items():
        if filename == 'demo_only_scrapy_cfg' or not alert:
            alert = 'the first selected node returned status'
        upload_file_deploy_multinode(filename='%s.zip' % filename, project=filename, alert=alert, fail=True)


def test_deploy_xhr(app, client):
    upload_file_deploy(app, client, filename='demo.egg', project=cst.PROJECT, redirect_project=cst.PROJECT, multinode=False)
    kws = dict(
        node=1,
        eggname='%s_%s_from_file_demo.egg' % (cst.PROJECT, cst.VERSION),
        project=cst.PROJECT,
        version=cst.VERSION
    )
    req(app, client, view='deploy.xhr', kws=kws, jskws=dict(status=cst.OK, project=cst.PROJECT))



# coding: utf-8
from flask import url_for

from tests.utils import cst, req, sleep, upload_file_deploy


jobid = ''


def test_index(app, client):
    with app.test_request_context():
        req(app, client, view='index', kws=dict(ui='mobile'),
            location=url_for('jobs', node=1, ui='mobile'))


def test_jobs(app, client):
    req(app, client, view='jobs', kws=dict(node=1, ui='mobile'),
        ins='jobs - ScrapydWeb - mobileui', mobileui=True)


def test_api_start(app, client):
    global jobid
    upload_file_deploy(app, client, filename='demo.egg', project=cst.PROJECT, redirect_project=cst.PROJECT)

    __, js = req(app, client, view='api',
                 kws=dict(node=1, opt='start', project=cst.PROJECT, version_spider_job=cst.SPIDER),
                 jskws=dict(status=cst.OK), jskeys='jobid')
    jobid = js['jobid']


# {'prevstate': running, 'status': 'ok',
# 'status_code': 200, 'url': 'http://127.0.0.1:6800/cancel.json'}
# In demo.egg: 'CONCURRENT_REQUESTS': 1, 'DOWNLOAD_DELAY': 10
def test_api_stop(app, client):
    sleep()
    req(app, client, view='api', kws=dict(node=1, opt='stop', project=cst.PROJECT, version_spider_job=jobid),
        jskws=dict(status=cst.OK, prevstate='running'), nos='times')


def test_api_forcestop(app, client):
    sleep(5)
    req(app, client, view='api', kws=dict(node=1, opt='forcestop', project=cst.PROJECT, version_spider_job=jobid),
        jskws=dict(status=cst.OK, prevstate=None, times=2))


def test_log_utf8(app, client):
    req(app, client, view='log',
        kws=dict(node=1, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER, job=jobid, ui='mobile'),
        ins='PROJECT (%s)' % cst.PROJECT, mobileui=True)


def test_log_stats(app, client):
    req(app, client, view='log',
        kws=dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=jobid, ui='mobile'),
        ins='current_time', mobileui=True)
